#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import re

from simplejson import dumps

from thingsboard_gateway.connectors.ftp.ftp_converter import FTPConverter
from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class FTPUplinkConverter(FTPConverter):
    def __init__(self, config, logger):
        self._log = logger
        self.__config = config
        self.__data_types = {"attributes": "attributes", "timeseries": "telemetry"}

    def _get_device_report_strategy(self, device_name):
        device_report_strategy = None
        try:
            device_report_strategy = ReportStrategyConfig(self.__config.get(REPORT_STRATEGY_PARAMETER))
        except ValueError as e:
            self._log.trace("Report strategy config is not specified for device %s: %s", device_name, e)

        return device_report_strategy

    def _get_required_data(self, left_symbol, right_symbol):
        device_name = None
        device_type = None
        get_device_name_from_data = False
        get_device_type_from_data = False

        if left_symbol in self.__config['devicePatternName'] and right_symbol in self.__config['devicePatternName']:
            get_device_name_from_data = True
        else:
            device_name = self.__config['devicePatternName']
        if left_symbol in self.__config['devicePatternType'] and right_symbol in self.__config['devicePatternType']:
            get_device_type_from_data = True
        else:
            device_type = self.__config['devicePatternType']

        return device_name, device_type, get_device_name_from_data, get_device_type_from_data

    def _convert_table_view_data(self, config, data):
        device_name, device_type, get_device_name_from_data, get_device_type_from_data = self._get_required_data('${',
                                                                                                                 '}')

        device_report_strategy = self._get_device_report_strategy(device_name)

        converted_data = ConvertedData(device_name=device_name,
                                        device_type=device_type)

        try:
            arr = data.split(self.__config.get('delimiter',';'))
            for data_type in self.__data_types:
                ts = None
                old_ts = None
                if data_type == 'timeseries':
                    ts = self._retrieve_ts_for_sliced_or_table(self.__config[data_type], arr, config['headers'])
                    old_ts = ts
                for information in self.__config[data_type]:
                    try:

                        key_index = information['key']
                        val_index = information['value']

                        if '${' in information['key'] and '}' in information['key']:
                            key_index = config['headers'].index(re.sub(r'[^\w]', '', information['key']))

                        if '${' in information['value'] and '}' in information['value']:
                            val_index = config['headers'].index(re.sub(r'[^\w]', '', information['value']))

                        key = arr[key_index] if isinstance(key_index, int) else key_index
                        value = arr[val_index] if isinstance(val_index, int) else val_index
                        if key == 'ts' and data_type == 'timeseries':
                            continue
                        datapoint_key = TBUtility.convert_key_to_datapoint_key(key, device_report_strategy,
                                                                               information, self._log)

                        if data_type == 'attributes':
                            converted_data.add_to_attributes(datapoint_key, value)
                        else:
                            if information.get('tsField'):
                                old_ts = ts
                                try:
                                    data_to_retrieve_ts = {}
                                    for index, item in enumerate(config['headers']):
                                       data_to_retrieve_ts[item] = arr[index]
                                    ts = TBUtility.resolve_different_ts_formats(data=data_to_retrieve_ts, config=information, logger=self._log)
                                except Exception as e:
                                    self._log.error('Error while retrieving timestamp for key %s: %s', key, e)
                                    ts = old_ts
                            else:
                                ts = old_ts
                            telemetry_entry = TelemetryEntry({datapoint_key: value}, ts)
                            converted_data.add_to_telemetry(telemetry_entry)

                        if get_device_name_from_data:
                            index = config['headers'].index(re.sub(r'[^\w]', '', self.__config['devicePatternName']))
                            converted_data.device_name = arr[index]
                        if get_device_type_from_data:
                            index = config['headers'].index(re.sub(r'[^\w]', '', self.__config['devicePatternType']))
                            converted_data.device_type = arr[index]
                    except Exception as e:
                        StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                        self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(information),
                                        data, e)

        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config), data,
                            e)

        StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)
        return converted_data

    @staticmethod
    def _get_key_or_value(key, arr):
        if '[' in key and ']' in key:
            split_val_arr = key[1:-1].split(':')
            first_val_index = split_val_arr[0] or 0
            last_val_index = split_val_arr[1] or len(arr)

            return arr[0][int(first_val_index):int(last_val_index)]
        else:
            return key

    def _convert_slices_view_data(self, data):
        device_name, device_type, get_device_name_from_data, get_device_type_from_data = self._get_required_data('[',
                                                                                                                 ']')

        device_report_strategy = self._get_device_report_strategy(device_name)

        converted_data = ConvertedData(device_name=device_name,
                                       device_type=device_type)
        arr = data.split(self.__config.get('delimiter', ','))

        try:
            for data_type in self.__data_types:
                ts = None
                if data_type == 'timeseries':
                    ts = self._retrieve_ts_for_sliced_or_table(self.__config[data_type], arr)
                for information in self.__config[data_type]:

                    try:

                        val = self._get_key_or_value(information['value'], arr)
                        key = self._get_key_or_value(information['key'], arr)
                        if key == 'ts' and data_type == 'timeseries':
                            continue

                        datapoint_key = TBUtility.convert_key_to_datapoint_key(key, device_report_strategy,
                                                                               information, self._log)
                        if data_type == 'attributes':
                            converted_data.add_to_attributes(datapoint_key, val)
                        else:
                            telemetry_entry = TelemetryEntry({datapoint_key: val}, ts)
                            converted_data.add_to_telemetry(telemetry_entry)

                        if get_device_name_from_data:
                            if self.__config['devicePatternName'] == information['value']:
                                converted_data.device_name = val
                        if get_device_type_from_data:
                            if self.__config['devicePatternType'] == information['value']:
                                converted_data.device_type = val
                    except Exception as e:
                        StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                        self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(information),
                                        data, e)
        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config), data,
                            e)

        StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)
        return converted_data

    def _retrieve_ts_for_sliced_or_table(self, config, data, headers=[]):
        for config_object in config:
            if config_object.get('key') == 'ts' or config_object.get('key') == 'timestamp':
                value = config_object['value']
                if '${' in config_object.get('value') and '}' in config_object.get('value') and headers:
                    value = headers.index(re.sub(r'[^\w]', '', config_object['value']))
                    return int(data[value])
                return self._get_key_or_value(value, data)
        return None

    def _get_device_name(self, data):
        device_name = ''

        try:
            if self.__config.get("devicePatternName") is not None:
                device_name_tags = TBUtility.get_values(self.__config.get("devicePatternName"), data,
                                                        get_tag=True)
                device_name_values = TBUtility.get_values(self.__config.get("devicePatternName"), data,
                                                          expression_instead_none=True)

                device_name = self.__config.get("devicePatternName")
                for (device_name_tag, device_name_value) in zip(device_name_tags, device_name_values):
                    is_valid_key = "${" in self.__config.get("devicePatternName") and "}" in \
                                   self.__config.get("devicePatternName")
                    device_name = device_name.replace('${' + str(device_name_tag) + '}',
                                                                                  str(device_name_value)) \
                        if is_valid_key else device_name_tag
        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config), data,
                            e)

        return device_name

    def _get_device_type(self, data):
        device_type = ''

        try:
            if self.__config.get("devicePatternType") is not None:
                device_type_tags = TBUtility.get_values(self.__config.get("devicePatternType"), data,
                                                        get_tag=True)
                device_type_values = TBUtility.get_values(self.__config.get("devicePatternType"), data,
                                                          expression_instead_none=True)
                device_type = self.__config.get("devicePatternType")

                for (device_type_tag, device_type_value) in zip(device_type_tags, device_type_values):
                    is_valid_key = "${" in self.__config.get("devicePatternType") and "}" in \
                                   self.__config.get("devicePatternType")
                    device_type = device_type.replace('${' + str(device_type_tag) + '}',
                                                                                  str(device_type_value)) \
                        if is_valid_key else device_type_tag
        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config), data,
                            e)

        return device_type

    def _convert_json_file(self, data):
        device_name = self._get_device_name(data)
        if not device_name:
            self._log.error("The expression for looking \"deviceName\" not found in config %s",
                            dumps(self.__config))

        device_type = self._get_device_type(data)
        if not device_type:
            self._log.error("The expression for looking \"deviceType\" not found in config %s",
                            dumps(self.__config))

        device_report_strategy = self._get_device_report_strategy(device_name)

        converted_data = ConvertedData(device_name=device_name,
                                       device_type=device_type)

        try:
            for datatype in self.__data_types:
                for datatype_config in self.__config.get(datatype, []):
                    try:
                        values = TBUtility.get_values(datatype_config["value"], data, datatype_config["type"],
                                                      expression_instead_none=True)
                        values_tags = TBUtility.get_values(datatype_config["value"], data, datatype_config["type"],
                                                           get_tag=True)

                        keys = TBUtility.get_values(datatype_config["key"], data, datatype_config["type"],
                                                    expression_instead_none=True)
                        keys_tags = TBUtility.get_values(datatype_config["key"], data, get_tag=True)

                        full_key = datatype_config["key"]
                        for (key, key_tag) in zip(keys, keys_tags):
                            is_valid_key = "${" in datatype_config["key"] and "}" in datatype_config["key"]
                            full_key = full_key.replace('${' + str(key_tag) + '}',
                                                        str(key)) if is_valid_key else key_tag
                        if full_key == 'ts' and datatype == 'timeseries':
                            continue

                        full_value = datatype_config["value"]
                        for (value, value_tag) in zip(values, values_tags):
                            is_valid_value = "${" in datatype_config["value"] and "}" in datatype_config["value"]

                            full_value = full_value.replace('${' + str(value_tag) + '}',
                                                            str(value)) if is_valid_value else str(value)
                        datapoint_key = TBUtility.convert_key_to_datapoint_key(full_key, device_report_strategy,
                                                                               datatype_config, self._log)
                        if datatype == 'timeseries':
                            ts = None
                            if data.get("ts") is not None or data.get("timestamp") is not None:
                                ts = data.get('ts', data.get('timestamp'))
                            telemetry_entry = TelemetryEntry({datapoint_key: full_value}, ts)
                            converted_data.add_to_telemetry(telemetry_entry)
                        else:
                            converted_data.add_to_attributes(datapoint_key, full_value)
                    except Exception as e:
                        StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                        self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s',
                                        dumps(datatype_config), data, e)
        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config),
                            str(data), e)

        StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)
        return converted_data

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, config, data):
        if data:
            if config['file_ext'] == 'csv' or (
                    config['file_ext'] == 'txt' and self.__config['txt_file_data_view'] == 'TABLE'):
                return self._convert_table_view_data(config, data)
            elif config['file_ext'] == 'txt' and self.__config['txt_file_data_view'] == 'SLICED':
                return self._convert_slices_view_data(data)
            elif config['file_ext'] == 'json':
                return self._convert_json_file(data)
            else:
                StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                raise Exception('Incorrect file extension or file data view mode')



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import asyncio
from threading import Thread
from time import sleep, monotonic
from typing import TYPE_CHECKING, Dict, Union

from _asyncio import Future

from thingsboard_gateway.connectors.modbus.constants import PymodbusDefaults
from thingsboard_gateway.connectors.modbus.bytes_modbus_downlink_converter import BytesModbusDownlinkConverter
from thingsboard_gateway.connectors.modbus.bytes_modbus_uplink_converter import BytesModbusUplinkConverter
from thingsboard_gateway.connectors.modbus.constants import (
    BAUDRATE_PARAMETER,
    BYTE_ORDER_PARAMETER,
    BYTESIZE_PARAMETER,
    CONNECT_ATTEMPT_COUNT_PARAMETER,
    CONNECT_ATTEMPT_TIME_MS_PARAMETER,
    HOST_PARAMETER,
    METHOD_PARAMETER,
    PARITY_PARAMETER,
    PORT_PARAMETER,
    REPACK_PARAMETER,
    RETRIES_PARAMETER,
    RPC_SECTION,
    SERIAL_CONNECTION_TYPE_PARAMETER,
    STOPBITS_PARAMETER,
    TIMEOUT_PARAMETER,
    UNIT_ID_PARAMETER,
    WAIT_AFTER_FAILED_ATTEMPTS_MS_PARAMETER,
    WORD_ORDER_PARAMETER,
    DELAY_BETWEEN_REQUESTS_MS_PARAMETER,
    TAG_PARAMETER
)
from thingsboard_gateway.connectors.modbus.entities.bytes_uplink_converter_config import BytesUplinkConverterConfig
from thingsboard_gateway.connectors.modbus.modbus_converter import ModbusConverter
from thingsboard_gateway.gateway.constants import (
    DEVICE_NAME_PARAMETER,
    DEVICE_TYPE_PARAMETER,
    TYPE_PARAMETER,
    UPLINK_PREFIX,
    CONVERTER_PARAMETER,
    DOWNLINK_PREFIX
)
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader

if TYPE_CHECKING:
    from thingsboard_gateway.connectors.modbus.modbus_connector import ModbusConnector
    from thingsboard_gateway.connectors.modbus.entities.master import Master


class Slave(Thread):
    def __init__(self, connector: 'ModbusConnector', logger, config):
        super().__init__()
        self.daemon = True
        self.stopped = False
        self._log = logger
        self.connector = connector
        self.type = config.get(TYPE_PARAMETER, 'tcp').lower()

        if self.type == SERIAL_CONNECTION_TYPE_PARAMETER:
            self.name = "Modbus slave processor for unit " + \
                str(config[UNIT_ID_PARAMETER]) + " on port " + \
                str(config[PORT_PARAMETER]) + ' ' + config[DEVICE_NAME_PARAMETER]
        else:
            self.name = "Modbus slave processor for unit " + \
                str(config[UNIT_ID_PARAMETER]) + " on host " + str(config.get('host')) + \
                ":" + str(config[PORT_PARAMETER]) + ' ' + config[DEVICE_NAME_PARAMETER]

        self.callback = connector.callback

        self.unit_id = config[UNIT_ID_PARAMETER]
        self.host = config.get(HOST_PARAMETER)
        self.port = config[PORT_PARAMETER]
        self.method = config[METHOD_PARAMETER].upper()
        self.tls = config.get('tls', {})
        self.timeout = config.get(TIMEOUT_PARAMETER, 30)
        self.retries = config.get(RETRIES_PARAMETER, 3)
        self.baudrate = config.get(BAUDRATE_PARAMETER, 19200)
        self.stopbits = config.get(STOPBITS_PARAMETER, PymodbusDefaults.Stopbits)
        self.bytesize = config.get(BYTESIZE_PARAMETER, PymodbusDefaults.Bytesize)
        self.parity = config.get(PARITY_PARAMETER, PymodbusDefaults.Parity)
        self.repack = config.get(REPACK_PARAMETER, False)
        self.word_order = config.get(WORD_ORDER_PARAMETER, 'LITTLE').upper()
        self.byte_order = config.get(BYTE_ORDER_PARAMETER, 'LITTLE').upper()
        self.handle_local_echo = config.get('handleLocalEcho', False)

        self.attributes_updates_config = config.get('attributeUpdates', [])
        self.rpc_requests_config = config.get(RPC_SECTION, [])

        self.connect_attempt_time_ms = config.get(CONNECT_ATTEMPT_TIME_MS_PARAMETER, 500) \
            if config.get(CONNECT_ATTEMPT_TIME_MS_PARAMETER, 500) > 500 else 500
        self.wait_after_failed_attempts_ms = config.get(WAIT_AFTER_FAILED_ATTEMPTS_MS_PARAMETER, 30000) \
            if config.get(WAIT_AFTER_FAILED_ATTEMPTS_MS_PARAMETER, 30000) > 30000 else 30000
        self.connection_attempt = config.get(CONNECT_ATTEMPT_COUNT_PARAMETER, 2) \
            if config.get(CONNECT_ATTEMPT_COUNT_PARAMETER, 2) >= 2 else 2

        self.__delay_between_requests_ms = config.get(DELAY_BETWEEN_REQUESTS_MS_PARAMETER, 0)

        self.device_name = config[DEVICE_NAME_PARAMETER]
        self.device_type = config.get(DEVICE_TYPE_PARAMETER, 'default')

        self.poll_period = config['pollPeriod'] / 1000

        self.last_connect_time = 0
        self.last_polled_time = 0
        self.last_connection_attempt_time = 0
        self.connection_attempt_count = 0

        self.downlink_converter = self.__load_downlink_converter(config)

        self.uplink_converter_config = BytesUplinkConverterConfig(**config)
        self.uplink_converter = self.__load_uplink_converter(config)

        self.__master: 'Master' = None
        self.available_functions = None

        self.shared_attributes_keys = []
        for attr_config in self.attributes_updates_config:
            self.shared_attributes_keys.append(attr_config[TAG_PARAMETER])

    def __timer(self):
        if self.__master is not None:
            self.__send_callback(monotonic())
            next_poll_time = monotonic() + self.poll_period

            while not self.stopped and not self.connector.is_stopped():
                current_time = monotonic()
                if current_time >= next_poll_time:
                    self.__send_callback(current_time)
                    next_poll_time = current_time + self.poll_period

                sleep_time = max(0.0, next_poll_time - monotonic())
                sleep(sleep_time)

    def __send_callback(self, current_monotonic):
        self.last_polled_time = current_monotonic

        try:
            self.callback(self, self.connector.process_device_requests)
        except Exception as e:
            self._log.exception('Error sending slave callback: %s', e)

    def run(self):
        if self.uplink_converter_config.is_readable():
            self.__timer()

    def close(self, loop):
        future = asyncio.run_coroutine_threadsafe(self.disconnect(), loop)
        try:
            future.result(timeout=5)
        except Exception as e:
            self._log.error('Failed to disconnect from %s: %s', self, e)
        self.stopped = True

    def get_name(self):
        return self.device_name

    def __load_downlink_converter(self, config):
        return self.__load_converter(config, DOWNLINK_PREFIX, {})

    def __load_uplink_converter(self, config):
        return self.__load_converter(config, UPLINK_PREFIX, self.uplink_converter_config)

    def __load_converter(self, config, converter_type, converter_config: Union[Dict, BytesUplinkConverterConfig] = {}):
        try:
            if isinstance(config.get(converter_type + CONVERTER_PARAMETER), str):
                converter = TBModuleLoader.import_module(self.connector.connector_type,
                                                         config[converter_type + CONVERTER_PARAMETER])(
                    converter_config, self._log)
            elif isinstance(config.get(converter_type + CONVERTER_PARAMETER), ModbusConverter):
                converter = config[converter_type + CONVERTER_PARAMETER]
            else:
                if converter_type == DOWNLINK_PREFIX:
                    converter = BytesModbusDownlinkConverter(converter_config, self.connector.converter_log)
                else:
                    converter = BytesModbusUplinkConverter(converter_config, self.connector.converter_log)

            return converter
        except Exception as e:
            self._log.exception('Failed to load %s converter for %s slave: %s', converter_type, self.name, e)

    async def connect(self) -> bool:
        cur_time = monotonic() * 1000
        initial_connection = False
        if not self.master.connected():
            if (self.connection_attempt_count >= self.connection_attempt
                    and cur_time - self.last_connection_attempt_time >= self.wait_after_failed_attempts_ms):
                self.connection_attempt_count = 0

            while not self.master.connected() \
                    and self.connection_attempt_count < self.connection_attempt \
                    and (cur_time - self.last_connection_attempt_time >= self.connect_attempt_time_ms
                         or self.last_connection_attempt_time == 0):
                if self.stopped:
                    return False
                self.connection_attempt_count += 1
                self.last_connection_attempt_time = cur_time
                self._log.debug("Trying connect to %s", self)

                try:
                    await asyncio.wait_for(self.master.connect(), timeout=self.timeout)
                except asyncio.exceptions.TimeoutError:
                    self._log.warning("Timeout error while connecting to %s", self)

                if self.connection_attempt_count == self.connection_attempt:
                    self._log.warning("Maximum attempt count (%i) for device \"%s\" - encountered.",
                                      self.connection_attempt,
                                      self)
                    return False
                initial_connection = True

        if self.connection_attempt_count >= 0 and self.master.connected():
            self.connection_attempt_count = 0
            self.last_connection_attempt_time = cur_time
            if initial_connection:
                self._log.info("Connected to %s", self)
            return True
        else:
            return False

    async def disconnect(self):
        if self.master.connected():
            await self.master.close()

    @property
    def master(self):
        return self.__master

    @master.setter
    def master(self, master):
        if master is not None:
            self.__master = master
            self.available_functions = self.__master.get_available_functions()
        else:
            self._log.warning('Master is not set for slave %s', self.device_name)

    async def read(self, function_code, address, objects_count):
        self._log.debug('Reading %s registers from address %s with function code %s', objects_count, address,
                        function_code)

        result = await self.__read(function_code, address, objects_count)

        StatisticsService.count_connector_message(self.connector.get_name(),
                                                  stat_parameter_name='connectorMsgsReceived')
        StatisticsService.count_connector_bytes(self.connector.get_name(), result,
                                                stat_parameter_name='connectorBytesReceived')

        return result

    async def __read(self, function_code, address, objects_count):
        result = None

        try:
            time_to_sleep = self.__master.get_time_to_pass_delay_between_requests(self.__delay_between_requests_ms)
            if time_to_sleep > 0:
                await asyncio.sleep(time_to_sleep / 1000)
            result = await self.available_functions[function_code](address=address,
                                                                   count=objects_count,
                                                                   unit_id=self.unit_id)
        except KeyError:
            self._log.error('Unknown Modbus function with code: %s', function_code)

        self._log.debug("Read with result: %s", str(result))
        return result

    async def write(self, function_code, address, value):
        self._log.debug('Write %s value to address %s with function code %s', value, address, function_code)
        result = await self.__write(function_code, address, value)

        if isinstance(result, Future):
            if not result.done():
                result = await result
            if result.exception() is not None:
                raise result.exception()

        StatisticsService.count_connector_message(self.connector.get_name(),
                                                  stat_parameter_name='connectorMsgsReceived')
        StatisticsService.count_connector_bytes(self.connector.get_name(), result,
                                                stat_parameter_name='connectorBytesReceived')

        return result

    async def __write(self, function_code, address, value):
        result = None

        try:
            time_to_sleep = self.__master.get_time_to_pass_delay_between_requests(self.__delay_between_requests_ms)
            if time_to_sleep > 0:
                await asyncio.sleep(time_to_sleep / 1000)
            if function_code in (5, 6):
                result = await self.available_functions[function_code](address=address, value=value,
                                                                       unit_id=self.unit_id)
            elif function_code in (15, 16):
                if not isinstance(value, list):
                    value = [value]

                result = await self.available_functions[function_code](address=address, values=value,
                                                                       unit_id=self.unit_id)
            else:
                self._log.error("Unknown Modbus function with code: %s", function_code)
        except Exception as e:
            self._log.error("Failed to write with function code %s: %s", function_code, e)
            future = asyncio.Future()
            future.set_exception(e)
            return future

        self._log.debug("Write with result: %s", str(result))
        return result

    def is_connected_to_platform(self):
        return self.last_connect_time != 0 and monotonic() - self.last_connect_time < 10

    def get_device_rpc_config(self, rpc_method):
        if isinstance(self.rpc_requests_config, dict):
            return self.rpc_requests_config.get(rpc_method)
        elif isinstance(self.rpc_requests_config, list):
            for rpc_command_config in self.rpc_requests_config:
                if rpc_command_config[TAG_PARAMETER] == rpc_method:
                    return rpc_command_config
        else:
            return None

    def __str__(self):
        return f'{self.device_name}'



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from pymodbus.datastore.context import ModbusSlaveContext
from pymodbus.datastore.store import ModbusSequentialDataBlock


class SlaveContext(ModbusSlaveContext):
    """
    Monkey-patched version of ModbusSlaveContext to allow initialization
    with custom data blocks for each type of data.
    Can be removed when the issue with ModbusSlaveContext is fixed in pymodbus.
    """

    def __init__(self, *_args, di=None, co=None, ir=None, hr=None):
        self.store = {}
        self.store["d"] = di if di is not None else ModbusSequentialDataBlock.create()
        self.store["c"] = co if co is not None else ModbusSequentialDataBlock.create()
        self.store["i"] = ir if ir is not None else ModbusSequentialDataBlock.create()
        self.store["h"] = hr if hr is not None else ModbusSequentialDataBlock.create()



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from typing import Dict, Any, Union

from thingsboard_gateway.gateway.entities.datapoint_key import DatapointKey


class Attributes:
    def __init__(self, values: Dict[DatapointKey, Any] = None):
        self.values: Dict[DatapointKey, Any] = values or {}

    def __str__(self):
        return f"Attributes(values={self.values})"

    def __hash__(self):
        return hash(tuple(self.values.items()))

    def to_dict(self) -> Dict[str, Any]:
        return {key.key if isinstance(key, DatapointKey) else key: value for key, value in self.values.items()}

    def __getitem__(self, item: DatapointKey) -> Any:
        return self.values[item]

    def __iter__(self):
        return iter(self.values)

    def __setitem__(self, key: DatapointKey, value):
        self.values[key] = value

    def __len__(self):
        return len(self.values)

    def update(self, attributes: Union[Dict[DatapointKey, Any], 'Attributes']):
        self.values.update(attributes if isinstance(attributes, dict) else attributes.values)

    def items(self):
        return self.values.items()



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.


from math import ceil
from struct import pack, unpack

from thingsboard_gateway.connectors.converter import Converter


class SerialDownlinkConverter(Converter):
    """
    Downlink converter is used to convert data that can be sent to device.
    Such as we create downlink converter for each configured device,
    this converter is used to convert data that can be sent to only one device.
    Because config, that we passed to init method, is device specific.
    If your connector can handle multiple devices, you can create one converter for all devices.
    """

    def __init__(self, config, logger):
        self._log = logger
        self.__config = config

    def convert(self, config, data) -> bytes:
        """Method to convert data that can be send to serial port."""
        self._log.debug("Data to convert: %s", data)
        byteorder = self.__config.get('byteorder', 'big').lower()
        if data is None:
            return None
        type_ = config.get("type")
        if type_ == "int":
            length = ceil(data.bit_length() / 8)
            return data.to_bytes(length, byteorder=byteorder)
        elif type_ == "float" or type_ == "double":
            fmt_single_precision = ('>' if byteorder == 'big' else '<') + 'f'
            single_precision_bytes = pack(fmt_single_precision, data)
            if unpack(fmt_single_precision, single_precision_bytes)[0] == data:
                return single_precision_bytes
            fmt_double_precision = ('>' if byteorder == 'big' else '<') + 'd'
            return pack(fmt_double_precision, data)
        return data.encode("UTF-8")



from copy import deepcopy


class BackwardCompatibilityAdapter:
    def __init__(self, config):
        self._config = deepcopy(config)

    @staticmethod
    def is_old_config(config):
        return config.get('socket') is None

    def convert(self):
        socket_type = self._config.pop('type', 'TCP')
        address = self._config.pop('address', '127.0.0.1')
        port = self._config.pop('port', 50000)
        buffer_size = self._config.pop('bufferSize', 1024)

        self._config['socket'] = {
            'type': socket_type,
            'address': address,
            'port': port,
            'bufferSize': buffer_size
        }

        for device in self._config.get('devices', []):
            if device.get('addressFilter'):
                address_filter = device.pop('addressFilter')
                device['address'] = address_filter

            for attribute_requests in device.get('attributeRequests', []):
                attribute_requests['requestExpressionSource'] = 'expression'
                attribute_requests['attributeNameExpressionSource'] = 'expression'

        return self._config



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import logging
import logging.handlers
import threading
from queue import Full, Queue, Empty
from sys import stdout
from time import time, sleep
from typing import TYPE_CHECKING

from thingsboard_gateway.tb_utility.tb_logger import TbLogger
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
if TYPE_CHECKING:
    from thingsboard_gateway.gateway.tb_gateway_service import TBGatewayService

logging.setLoggerClass(TbLogger)


class TBRemoteLoggerHandler(logging.Handler):
    LOGGER_NAME_TO_ATTRIBUTE_NAME = {
        'service': 'SERVICE_LOGS',
        'extension': 'EXTENSION_LOGS',
        'tb_connection': 'CONNECTION_LOGS',
        'storage': 'STORAGE_LOGS',
    }

    def __init__(self, gateway: 'TBGatewayService'):
        logging.setLoggerClass(TbLogger)
        self.current_log_level = self.get_logger_level_id('TRACE')
        super().__init__(self.current_log_level)
        self.setLevel(self.current_log_level)
        self.__gateway = gateway
        self.activated = False
        self.__loggers_lock = threading.Lock()

        self._max_message_count_batch = 20
        self._logs_queue = Queue(1000)

        self._send_logs_thread = None

        self.setFormatter(logging.Formatter('%(asctime)s - |%(levelname)s|<%(threadName)s> [%(filename)s] - %(module)s %(funcName)s - %(lineno)d - %(message)s')) # noqa
        self.loggers = {}
        for logger in TBRemoteLoggerHandler.LOGGER_NAME_TO_ATTRIBUTE_NAME.keys():
            self.add_logger(logger, 100)
        self.activate()

    def activate_remote_logging_for_level(self, log_level_id):
        for logger_name in TBRemoteLoggerHandler.LOGGER_NAME_TO_ATTRIBUTE_NAME.keys():
            logger, _ = self.loggers.get(logger_name)
            if logger:
                self.loggers[logger_name] = logger, log_level_id

    def get_logger(self, name):
        return self.loggers.get(name)[0]

    def add_logger(self, name, remote_logging_level):
        log = logging.getLogger(name)
        with self.__loggers_lock:
            self.loggers[name] = log, self.get_logger_level_id(remote_logging_level)

    def update_logger(self, name, remote_logging_level):
        if name in self.loggers:
            with self.__loggers_lock:
                self.loggers[name] = self.loggers[name][0], self.get_logger_level_id(remote_logging_level)

    def remove_logger(self, name):
        if name in self.loggers:
            with self.__loggers_lock:
                self.loggers.pop(name)

    def _send_logs(self):
        while self.activated and not self.__gateway.stopped:
            try:
                if self.__gateway.tb_client is None or not self.__gateway.tb_client.is_connected():
                    sleep(1)
                    continue
                logs_for_sending_list = []
                log_msg = self._logs_queue.get_nowait()

                count = 1
                while count <= self._max_message_count_batch:
                    try:
                        if self.__gateway.tb_client is None or not self.__gateway.tb_client.is_connected():
                            sleep(1)
                            continue
                        if log_msg is None:
                            log_msg = self._logs_queue.get_nowait()

                        logs_msg_size = TBUtility.get_data_size(log_msg)
                        if logs_msg_size > self.__gateway.get_max_payload_size_bytes():
                            print(f'Too big LOG message size to send ({logs_msg_size}). Skipping...')
                            print(log_msg)
                            continue

                        if TBUtility.get_data_size(logs_for_sending_list) + logs_msg_size > self.__gateway.get_max_payload_size_bytes(): # noqa
                            self.__gateway.send_telemetry(logs_for_sending_list)
                            logs_for_sending_list = [log_msg]
                        else:
                            logs_for_sending_list.append(log_msg)
                        log_msg = None
                        count += 1
                    except Empty:
                        break

                if logs_for_sending_list and not self.__gateway.stopped:
                    self.__gateway.send_telemetry(logs_for_sending_list)
            except (TimeoutError, Empty):
                self.__gateway.stop_event.wait(.1)
            except Exception as e:
                log = TbLogger('service')
                log.debug("Exception while sending logs.", exc_info=e)

    def activate(self):
        if not self.activated:
            self.activated = True
            self._send_logs_thread = threading.Thread(target=self._send_logs, name='Logs Sending Thread', daemon=True)
            self._send_logs_thread.start()

    def handle(self, record):
        if self.activated and not self.__gateway.stopped:
            logger, remote_logging_level = self.loggers.get(record.name)
            if record.levelno < remote_logging_level:
                return  # Remote logging level set higher than record level

            name = logger.name

            if name:
                record = self.formatter.format(record)
                try:
                    telemetry_key = self.LOGGER_NAME_TO_ATTRIBUTE_NAME[name]
                except KeyError:
                    telemetry_key = name + '_LOGS'

                log_msg = {'ts': int(time() * 1000), 'values': {telemetry_key: record}}

                if telemetry_key in self.LOGGER_NAME_TO_ATTRIBUTE_NAME.values():
                    log_msg['values']['LOGS'] = record
                try:
                    self._logs_queue.put_nowait(log_msg)
                except Full:
                    print(f"Logs queue is full. Skipping log message: {log_msg}")
                except Exception as e:
                    print(f"Exception while putting log message to queue: {str(e)}")

    def deactivate(self):
        self.activated = False
        try:
            self._send_logs_thread.join()
        except Exception as e:
            log = TbLogger('service')
            log.debug("Exception while joining logs sending thread.", exc_info=e)

    @staticmethod
    def set_default_handler():
        logging.setLoggerClass(TbLogger)
        for logger_name in TBRemoteLoggerHandler.LOGGER_NAME_TO_ATTRIBUTE_NAME.keys():
            logger = logging.getLogger(logger_name)
            handler = logging.StreamHandler(stdout)
            handler.setFormatter(logging.Formatter('[STREAM ONLY] %(asctime)s,%(msecs)03d - %(levelname)s - [%(filename)s] - %(module)s - %(lineno)d - %(message)s')) # noqa
            logger.addHandler(handler)

    @staticmethod
    def get_logger_level_id(log_level):
        if isinstance(log_level, str):
            return 100 if log_level == 'NONE' else logging._nameToLevel.get(log_level, 100)
        return log_level



#      Copyright 2025. ThingsBoard
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from thingsboard_gateway.gateway.constants import *  # noqa


class PymodbusDefaults:
    TcpPort = 502
    TlsPort = 802
    UdpPort = 502
    Backoff = 0.3
    CloseCommOnError = False
    HandleLocalEcho = False
    Retries = 3
    RetryOnEmpty = False
    RetryOnInvalid = False
    Timeout = 3
    Reconnects = 0
    TransactionId = 0
    Strict = True
    ProtocolId = 0
    Slave = 0x00
    Baudrate = 19200
    Parity = "N"
    Bytesize = 8
    Stopbits = 1
    ZeroMode = False
    IgnoreMissingSlaves = False
    ReadSize = 1024
    BroadcastEnable = False
    ReconnectDelay = 1000 * 60 * 5
    Count = 1


# Connector constants

LAST_PREFIX = "last_"
NEXT_PREFIX = "next_"

CHECK_POSTFIX = "_check"
POLL_PERIOD_POSTFIX = "PollPeriod"

TIMESERIES_PARAMETER = "timeseries"

MASTER_PARAMETER = "master"
AVAILABLE_FUNCTIONS_PARAMETER = "available_functions"

CONNECTION_ATTEMPT_PARAMETER = "connection_attempt"
LAST_CONNECTION_ATTEMPT_TIME_PARAMETER = "last_connection_attempt_time"

# Configuration parameters

RPC_SECTION = "rpc"
IDENTITY_SECTION = "identity"
SECURITY_SECTION = "security"

BYTE_ORDER_PARAMETER = "byteOrder"
WORD_ORDER_PARAMETER = "wordOrder"
CONNECT_ATTEMPT_COUNT_PARAMETER = "connectAttemptCount"
CONNECT_ATTEMPT_TIME_MS_PARAMETER = "connectAttemptTimeMs"
WAIT_AFTER_FAILED_ATTEMPTS_MS_PARAMETER = "waitAfterFailedAttemptsMs"

DELAY_BETWEEN_REQUESTS_MS_PARAMETER = "delayBetweenRequestsMs"

FUNCTION_CODE_PARAMETER = "functionCode"

ADDRESS_PARAMETER = "address"
OBJECTS_COUNT_PARAMETER = "objectsCount"

UNIT_ID_PARAMETER = "unitId"

HOST_PARAMETER = "host"
PORT_PARAMETER = "port"
BAUDRATE_PARAMETER = "baudrate"
TIMEOUT_PARAMETER = "timeout"
METHOD_PARAMETER = "method"
STOPBITS_PARAMETER = "stopbits"
BYTESIZE_PARAMETER = "bytesize"
PARITY_PARAMETER = "parity"
TYPE_PARAMETER = "type"
REPACK_PARAMETER = "repack"
SERIAL_CONNECTION_TYPE_PARAMETER = "serial"

RETRIES_PARAMETER = "retries"

PAYLOAD_PARAMETER = "payload"
TAG_PARAMETER = "tag"

COILS_INITIALIZER = "coils_initializer"
HOLDING_REGISTERS = "holding_registers"
INPUT_REGISTERS = "input_registers"
DISCRETE_INPUTS = "discrete_inputs"

FUNCTION_TYPE = {
    COILS_INITIALIZER: 'co',
    HOLDING_REGISTERS: 'hr',
    INPUT_REGISTERS: 'ir',
    DISCRETE_INPUTS: 'di'
}

FUNCTION_CODE_SLAVE_INITIALIZATION = {
    HOLDING_REGISTERS: (6, 16),
    COILS_INITIALIZER: (5, 15),
    INPUT_REGISTERS: (6, 16),
    DISCRETE_INPUTS: (5, 15)
}

FUNCTION_CODE_READ = {
    HOLDING_REGISTERS: 3,
    COILS_INITIALIZER: 1,
    INPUT_REGISTERS: 4,
    DISCRETE_INPUTS: 2
}

# Default values

TIMEOUT = 30



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

class UplinkConverterConfig:
    def __init__(self, config, device_info, device_details):
        self.__config = config

        self.device_details = device_details
        self.device_name = device_info.device_name
        self.device_type = device_info.device_type
        self.__objects_to_read = self.__get_objects_to_read()
        self.report_strategy = self.__config.get('reportStrategy', {})

    def __get_objects_to_read(self):
        objects_to_read = []

        for section in ('attributes', 'timeseries'):
            section_config = self.__config.get(section, [])

            if section_config != '*':
                for item in section_config:
                    is_local_discovery_config = list(filter(lambda key_value: key_value == '*', item.values()))
                    if len(is_local_discovery_config) == 0:
                        item = {**item, 'type': section}

                        if item.get('objectType') is None:
                            continue

                        if item['objectType'] == 'device':
                            item['objectId'] = self.device_details.object_id

                        objects_to_read.append(item)

        return objects_to_read

    @property
    def objects_to_read(self):
        return self.__objects_to_read



#      Copyright 2025. ThingsBoard
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from xknx import XKNX
from xknx.io.gateway_scanner import GatewayScanner


class GatewaysScanner(GatewayScanner):
    def __init__(self, scanner_config, logger):
        self.__log = logger
        timeout_in_seconds = int(scanner_config.get('scanPeriod', 3))
        stop_on_found = scanner_config.get('stopOnFound', False)

        self.__client = XKNX()
        super().__init__(xknx=self.__client,
                         timeout_in_seconds=timeout_in_seconds,
                         stop_on_found=stop_on_found)

    @staticmethod
    def is_configured(client_config):
        gateways_scanner_config = client_config.get('gatewaysScanner', {})
        return gateways_scanner_config and gateways_scanner_config.get('enabled', False)

    @staticmethod
    def get_tunneling_type(gateway):
        tunnelling = (
            "Secure"
            if gateway.tunnelling_requires_secure
            else "TCP"
            if gateway.supports_tunnelling_tcp
            else "UDP"
            if gateway.supports_tunnelling
            else "No"
        )

        return tunnelling

    @staticmethod
    def get_routing_type(gateway):
        routing = (
            "Secure"
            if gateway.routing_requires_secure
            else "Yes"
            if gateway.supports_routing
            else "No"
        )

        return routing

    async def scan(self):
        self.__log.info('Starting KNX gateways scan...')

        async for found_gateway in self.async_scan():
            self.__log.info(
                'Found KNX gateway:\n'
                '- %s %s\n'
                '  %s:%s\n'
                '  Tunneling: %s\n'
                '  Routing: %s',
                found_gateway.name,
                found_gateway.individual_address,
                found_gateway.ip_addr,
                found_gateway.port,
                self.get_tunneling_type(found_gateway),
                self.get_routing_type(found_gateway)
            )

        if not self.found_gateways:
            self.__log.info('No KNX gateways found')



#      Copyright 2025. ThingsBoard
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import asyncio
from queue import Empty, Queue
from re import fullmatch
from threading import Thread, Event
from time import sleep, monotonic
from typing import Tuple, Any

from thingsboard_gateway.gateway.constants import STATISTIC_MESSAGE_SENT_PARAMETER, RPC_DEFAULT_TIMEOUT
from thingsboard_gateway.gateway.statistics.decorators import CollectAllReceivedBytesStatistics
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.tb_utility.tb_logger import init_logger

try:
    from xknx import XKNX
except ImportError:
    print("xknx library not found")
    TBUtility.install_package("xknx")
    from xknx import XKNX

from thingsboard_gateway.connectors.knx.entities.client_config import ClientConfig
from thingsboard_gateway.connectors.knx.entities.gateways_scanner import GatewaysScanner
from thingsboard_gateway.connectors.knx.entities.device import Device

from xknx.tools import read_group_value, group_value_write


class KNXConnector(Connector, Thread):
    def __init__(self, gateway, config, connector_type):
        super().__init__()

        self.__gateway = gateway
        self._connector_type = connector_type

        self.__log = init_logger(self.__gateway, config['name'],
                                 config.get('logLevel', 'INFO'),
                                 enable_remote_logging=config.get('enableRemoteLogging', False),
                                 is_connector_logger=True)
        self.__converter_log = init_logger(self.__gateway, config['name'] + '_converter',
                                           config.get('logLevel', 'INFO'),
                                           enable_remote_logging=config.get('enableRemoteLogging', False),
                                           is_converter_logger=True, attr_name=config['name'])

        self.statistics = {'MessagesReceived': 0, 'MessagesSent': 0}

        self.__config = config
        self.__id = self.__config.get('id')
        self.__connected = False
        self.__stopped = Event()
        self.daemon = True

        self.__process_device_request_queue = Queue(-1)
        self.__data_to_convert_queue = Queue(-1)
        self.__data_to_save_queue = Queue(-1)

        self.__loop = self.__create_event_loop()

        self.__is_client_connected = asyncio.Event()
        self.__client = None

        self.__devices = []
        self.__add_devices()

        self.__log.info('Starting KNX Connector...')

    def __add_devices(self):
        for device_config in self.__config.get('devices', []):
            self.__devices.append(Device(self._connector_type,
                                         self.__is_client_connected,
                                         device_config,
                                         self.__process_device_request_queue,
                                         self.__converter_log))

    def __create_event_loop(self):
        try:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return loop
        except RuntimeError:
            return asyncio.get_event_loop()

    def close(self):
        self.__stopped.set()
        self.__connected = False

        self.__log.debug('Stopping %s...', self.get_name())

        self.__stop_devices()

        if TBUtility.while_thread_alive(self):
            self.__log.error("Failed to stop connector %s", self.get_name())
        else:
            self.__log.info('%s has been stopped', self.get_name())

        self.__converter_log.stop()
        self.__log.stop()

    def __stop_devices(self):
        for device in self.__devices:
            device.stop()

    def open(self):
        self.start()

    @staticmethod
    def telegram_received_cb(telegram):
        print("Telegram received: {0}".format(telegram))

    def run(self):
        try:
            self.__connected = True
            self.__loop.run_until_complete(self.__start())
        except asyncio.CancelledError as e:
            self.__log.debug('Task was cancelled due to connector stop: %s', e.__str__())
        except Exception as e:
            self.__log.exception(e)

    async def __start(self):
        client_config = self.__config.get('client', {})

        # check if gateways scanner is configured
        await self.__check_and_scan_gateways(client_config)

        self.__create_client(client_config)

        await asyncio.gather(self.__start_client_with_block(),
                             self.__process_device_request(),
                             self.__convert_data(),
                             self.__send_data())

    async def __check_and_scan_gateways(self, client_config):
        if GatewaysScanner.is_configured(client_config):
            gateway_scanner = GatewaysScanner(client_config, self.__log)
            await gateway_scanner.scan()

    async def __start_client_with_block(self):
        while not self.__stopped.is_set():
            try:
                await self.__client.start()
                self.__log.info('Connected to KNX bus')
                self.__is_client_connected.set()

                while not self.__stopped.is_set():
                    await asyncio.sleep(.02)
            except Exception as e:
                self.__log.error('Error connecting to KNX bus: %s', e.__str__())

        await self.__client.stop()
        self.__is_client_connected.clear()

    def __create_client(self, client_config):
        client_config = ClientConfig(client_config)
        self.__client = XKNX(**client_config.__dict__)

    async def __process_device_request(self):
        while not self.__stopped.is_set():
            try:
                device = self.__process_device_request_queue.get_nowait()

                if self.__client.connection_manager.connected:
                    responses = {}

                    for key, config in device.group_addresses_to_read.items():
                        try:
                            response = await read_group_value(self.__client, key, config.get('type'))

                            if response is not None:
                                self.__log.trace('Response from KNX bus: %s', response)
                                responses[key] = {
                                    'type': config['type'],
                                    'response': response,
                                    'keys': config['keys']
                                }
                            else:
                                self.__log.warning('No response from KNX bus for %s.', key)
                        except Exception as e:
                            self.__log.exception('Error processign %s request: %s', device.name, e)

                    if responses:
                        self.__data_to_convert_queue.put((device, responses))
                else:
                    self.__log.error('KNX bus is not connected')
            except Empty:
                await asyncio.sleep(.01)

    async def __convert_data(self):
        while not self.__stopped.is_set():
            try:
                device, data_to_convert = self.__data_to_convert_queue.get_nowait()
                converted_data = device.uplink_converter.convert(data_to_convert)

                if converted_data.telemetry_datapoints_count > 0 or converted_data.attributes_datapoints_count > 0:
                    self.__data_to_save_queue.put((device, converted_data))
            except Empty:
                await asyncio.sleep(.01)
            except Exception as e:
                self.__log.error('Error converting data: %s', e)

    async def __send_data(self):
        while not self.__stopped.is_set():
            try:
                device, data_to_save = self.__data_to_save_queue.get_nowait()
                self.__log.trace('%s data to save: %s', device, data_to_save)
                StatisticsService.count_connector_message(self.get_name(), stat_parameter_name='storageMsgPushed')
                self.__gateway.send_to_storage(self.get_name(), self.get_id(), data_to_save)
                self.statistics[STATISTIC_MESSAGE_SENT_PARAMETER] += 1
            except Empty:
                await asyncio.sleep(.01)
            except Exception as e:
                self.__log.error('Error saving data: %s', e)

    @CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def on_attributes_update(self, content):
        self.__log.debug('Received attribute update request: %s', content)

        try:
            attribute_request_config, value = self.__find_attribute_request_by_device_and_name(content)
            if not attribute_request_config:
                self.__log.error('No attribute request config found for device %s', content['device'])
                return

            result = self.__process_attribute_update(group_address=attribute_request_config['groupAddress'],
                                                     data_type=attribute_request_config.get('dataType'), value=value)

            if result.get("response").get("error"):
                self.__log.error('Failed to process on attribute update request with result: %s', result.get("response"))
                return

            self.__log.info('Successfully processed on attribute update request with result: %s', result.get("response"))


        except Exception as e:
            self.__log.error('Error processing attribute update request: %s', e)

    def __find_attribute_request_by_device_and_name(self, content):
        for attribute_request_config in self.__config.get('attributeUpdates', []):
            device_name_match = fullmatch(attribute_request_config['deviceNameFilter'], content['device'])

            attr_name_match_fitler = False
            value = None
            for attr_k, attr_v in content['data'].items():
                if attr_k == attribute_request_config['key']:
                    attr_name_match_fitler = True
                    value = attr_v

            if device_name_match and attr_name_match_fitler:
                return attribute_request_config, value

    def __process_attribute_update(self, group_address, data_type, value):
        try:
            if self.__client.connection_manager.connected:
                group_value_write(self.__client, group_address, value, data_type)
                return {"response":{"value": str(value)}}
            else:
                self.__log.error('KNX bus is not connected')
                return {"response": {"error": "KNX bus is not connected"}}

        except Exception as e:
            result = {"response": {"error": str(e)}}
            self.__log.debug('Error: %r', str(e), exc_info=True)
            return result

    @CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def server_side_rpc_handler(self, content):
        self.__log.debug('Received RPC request: %s', content)

        rpc_method_name = content.get('data', {}).get('method')
        if rpc_method_name is None:
            self.__log.error('Method name not found in RPC request: %r', content)
            return

        # check if RPC method is reserved get/set
        is_processed_reserved_rpc = self.__check_and_process_reserved_rpc(rpc_method_name, content)
        if is_processed_reserved_rpc:
            return

        rpc_config = self.__find_rpc_config_by_device_and_name(rpc_method_name, content)
        if not rpc_config:
            self.__log.error('No RPC config found')
            return

        self.__process_rpc(content, rpc_config)

    def __check_and_process_reserved_rpc(self, rpc_method_name, content):
        if rpc_method_name in ('get', 'set'):
            params = {}

            for param in content['data']['params'].split(';'):
                try:
                    (key, value) = param.split('=')
                except ValueError:
                    self.__log.trace('Invalid parameter: %s', param)
                    continue

                if key and value:
                    params[key] = value

            content['data'].pop('params', None)

            if params.get('value'):
                content['data']['params'] = params['value']

            self.__process_rpc(content, params)
            return True

    @staticmethod
    def __wait_task_with_timeout(task: asyncio.Task, timeout: float, poll_interval: float = 0.2) -> Tuple[bool, Any]:
        start_time = monotonic()
        while not task.done():
            sleep(poll_interval)
            current_time = monotonic()
            if current_time - start_time >= timeout:
                task.cancel()
                return False, None
        return True, task.result()

    def __process_rpc(self, content, rpc_config):
        try:
            value = content.get('data', {}).get('params')
            task = self.__create_task(self.__process_rpc_request, (rpc_config,), {
                'value': value})
            task_completed, result = self.__wait_task_with_timeout(task=task,
                                                                   timeout=content.get("timeout", RPC_DEFAULT_TIMEOUT),
                                                                   poll_interval=0.2)
            if not task_completed:
                self.__log.error('RPC request timed out')
                result = {"error": f"Timeout rpc has been reached for {content['device']}"}

            self.__log.info('Processed RPC request with result: %r', result)
            self.__gateway.send_rpc_reply(content['device'],
                                          req_id=content['data'].get('id'),
                                          content={'result': str(result.get('response'))})
        except Exception as e:
            self.__log.error('Error processing RPC request: %s', e)
            self.__gateway.send_rpc_reply(device=content['device'],
                                          req_id=content['data'].get('id'),
                                          content={'result': str(e)},
                                          success_sent=False)

    async def __process_rpc_request(self, config, value=None):
        if self.__client.connection_manager.connected:
            group_address = config['groupAddress']
            data_type = config.get('dataType')

            if value is not None or config['requestType'] == 'write':

                response = self.__process_group_value_write(group_address, value, data_type)
            else:
                response = await self.__read_group_value(group_address, data_type)
            return {"response": response}
        else:
            self.__log.error('KNX bus is not connected')
            return {"response": {"error": "KNX bus is not connected"}}

    def __process_group_value_write(self, group_address, value, data_type):
        try:
            group_value_write(self.__client, group_address, value, data_type)
            return {"value": str(value)}

        except Exception as e:
            result = {}
            self.__log.error('Error processing group value write: %s', str(e))
            self.__log.debug('Error: %r', str(e), exc_info=True)
            result['error'] = str(e)
            return result

    async def __read_group_value(self, group_address, data_type):
        try:
            result = await read_group_value(self.__client, group_address, data_type)
            return {"value": str(result)}

        except Exception as e:
            result = {}
            self.__log.error('Error processing group value read: %s', str(e))
            self.__log.debug('Error: %r', str(e), exc_info=True)
            result['error'] = str(e)
            return result

    def __find_rpc_config_by_device_and_name(self, rpc_method_name, content):
        device_name_match_filter = tuple(filter(lambda rpc_config:
                                                fullmatch(rpc_config['deviceNameFilter'], content['device']),
                                                self.__config.get('serverSideRpc', [])))

        rpc_method_name_filter = tuple(filter(lambda rpc_config: rpc_config['method'] == rpc_method_name,
                                              self.__config.get('serverSideRpc', [])))

        if len(device_name_match_filter) and len(rpc_method_name_filter):
            return rpc_method_name_filter[0]

    def __create_task(self, func, args, kwargs):
        task = self.__loop.create_task(func(*args, **kwargs))
        return task

    @property
    def connector_type(self):
        return self._connector_type

    def get_id(self):
        return self.__id

    def get_name(self):
        return self.name

    def get_type(self):
        return self._connector_type

    def get_config(self):
        return self.__config

    def is_connected(self):
        return self.__connected

    def is_stopped(self):
        return self.__stopped.is_set()



import logging
from os import sep
from os.path import isfile, exists
from re import compile, ASCII
from logging.handlers import TimedRotatingFileHandler as BaseTimedRotatingFileHandler
from os import environ
from pathlib import Path


class TimedRotatingFileHandler(BaseTimedRotatingFileHandler):
    DELIMITER_RE = compile(r'[\s-]+')
    CAMEL_BOUNDARY_RE = compile(r'(?<=[a-z0-9])([A-Z])')
    SPECIAL_CHAR_RE = compile(r'[^A-Za-z0-9_]')
    MULTI_UNDERSCORE_RE = compile(r'_+')

    def __init__(self, filename, when='h', interval=1, backupCount=0,
                 encoding=None, delay=False, utc=False, maxBytes=0):
        file_path = filename
        config_path_sep = sep if file_path.rfind(sep) != -1 else '/'
        config_path = environ.get('TB_GW_LOGS_PATH')
        original_log_filename = file_path.split(config_path_sep)[-1]

        final_filename = original_log_filename.replace('.log', '')
        final_filename = TimedRotatingFileHandler.to_snake_case(final_filename)
        final_filename += '.log'

        file_path = file_path.replace(original_log_filename, final_filename)

        if config_path:
            file_path = config_path + sep + final_filename

        if not Path(file_path).exists():
            with open(file_path, 'w'):
                pass

        super().__init__(filename=file_path, when=when, interval=interval, backupCount=backupCount,
                         encoding=encoding, delay=delay, utc=utc)

        self.maxBytes = maxBytes
        if self.maxBytes > 0:
            self.suffix = "%Y-%m-%d_%H-%M-%S"
            self.extMatch = compile(r"(?<!\d)\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}(?!\d)", ASCII)

    def shouldRolloverOnSize(self, record):
        if self.stream is None:
            return False

        if self.maxBytes > 0:
            pos = self.stream.tell()
            if not pos:
                return False
            msg = "%s\n" % self.format(record)
            if pos + len(msg) >= self.maxBytes:
                if exists(self.baseFilename) and not isfile(self.baseFilename):
                    return False
                return True

        return False

    def shouldRollover(self, record):
        return self.shouldRolloverOnSize(record) or super().shouldRollover(record)

    @staticmethod
    def get_connector_file_handler(file_name):
        return TimedRotatingFileHandler.__get_file_handler(file_name, 'connector')

    @staticmethod
    def get_converter_file_handler(file_name):
        return TimedRotatingFileHandler.__get_file_handler(file_name, 'converter')

    @staticmethod
    def get_time_rotating_file_handler_by_logger_name(logger_name):
        file_handler_filter = list(filter(lambda x: isinstance(x, TimedRotatingFileHandler),
                                          logging.getLogger(logger_name).handlers))
        if len(file_handler_filter):
            return file_handler_filter[0]

    @staticmethod
    def __get_file_handler(file_name, logger_name):
        file_handler = TimedRotatingFileHandler.get_time_rotating_file_handler_by_logger_name(logger_name)

        if not file_name.endswith('.log'):
            file_name = file_name + '.log'

        if file_handler:
            return TimedRotatingFileHandler.__create_file_handler_copy(file_handler, file_name)
        else:
            return TimedRotatingFileHandler.__create_default_file_handler(file_name)

    @staticmethod
    def __create_file_handler_copy(handler, file_name):
        file_name = f'{sep}'.join(handler.baseFilename.split(sep)[:-1]) + sep + file_name
        handler_copy = TimedRotatingFileHandler(file_name,
                                                when=handler.when,
                                                backupCount=handler.backupCount,
                                                interval=handler.interval,
                                                encoding=handler.encoding,
                                                delay=handler.delay,
                                                utc=handler.utc,
                                                maxBytes=handler.maxBytes)
        handler_copy.setFormatter(handler.formatter)
        return handler_copy

    @staticmethod
    def __create_default_file_handler(file_name):
        if not file_name.endswith('.log'):
            file_name = file_name + '.log'
        return TimedRotatingFileHandler(file_name)

    @staticmethod
    def to_snake_case(name: str) -> str:
        s = TimedRotatingFileHandler.DELIMITER_RE.sub('_', name)
        s = TimedRotatingFileHandler.CAMEL_BOUNDARY_RE.sub(r'_\1', s)
        s = s.lower()
        s = TimedRotatingFileHandler.SPECIAL_CHAR_RE.sub('', s)
        s = TimedRotatingFileHandler.MULTI_UNDERSCORE_RE.sub('_', s)
        return s.strip('_')



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from threading import RLock
from typing import Dict


class EventStorageFiles:
    def __init__(self, state_file, data_files: Dict[str, bool]):
        self.state_file = state_file
        self.data_files = data_files
        self.__data_files_lock = RLock()

    def get_state_file(self):
        return self.state_file

    def get_data_files(self):
        with self.__data_files_lock:
            return sorted(self.data_files.keys())

    def add_data_file(self, data_file):
        with self.__data_files_lock:
            self.data_files[data_file] = False

    def remove_data_file(self, data_file):
        with self.__data_files_lock:
            del self.data_files[data_file]

    def confirm_file_processed(self, data_file):
        with self.__data_files_lock:
            self.data_files[data_file] = True



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from enum import Enum

from thingsboard_gateway.gateway.constants import (
    DATA_PARAMETER,
    DEVICE_SECTION_PARAMETER,
    RPC_ID_PARAMETER,
    RPC_PARAMS_PARAMETER,
    RPC_METHOD_PARAMETER
)


class RPCType(Enum):
    CONNECTOR = 'CONNECTOR'
    DEVICE = 'DEVICE'
    RESERVED = 'RESERVED'


class RPCRequest:
    def __init__(self, content):
        if self.is_old_format_rpc_content(content):
            content = self._convert_old_format_rpc_content(content)

        self.rpc_type = self._get_rpc_type(content)
        self.method = None
        self.params = None
        self._value = None
        self.timeout = content.get('timeout', 5.0)
        self.id = content[DATA_PARAMETER].get(RPC_ID_PARAMETER)
        self.device_name = content.get(DEVICE_SECTION_PARAMETER)

        self._fill_rpc_request(content)

    @property
    def value(self):
        return {DATA_PARAMETER: {RPC_PARAMS_PARAMETER: self._value}}

    def is_old_format_rpc_content(self, content):
        return content.get(DATA_PARAMETER) is None

    @staticmethod
    def _convert_old_format_rpc_content(content):
        new_content = {}
        new_content[DATA_PARAMETER] = {RPC_PARAMS_PARAMETER: content[RPC_PARAMS_PARAMETER],
                                       RPC_METHOD_PARAMETER: content[RPC_METHOD_PARAMETER]}
        return new_content

    def _get_rpc_type(self, content):
        if self._is_connector_rpc(content):
            return RPCType.CONNECTOR
        elif self._is_reserved_rpc(content):
            return RPCType.RESERVED
        elif self._is_device_rpc(content):
            return RPCType.DEVICE
        else:
            raise ValueError('Unknown RPC type for content: %r' % content)

    @staticmethod
    def _is_connector_rpc(content) -> bool:
        try:
            (connector_type, _) = content[DATA_PARAMETER][RPC_METHOD_PARAMETER].split('_')
            if connector_type == 'modbus':
                return True
        except (IndexError, ValueError):
            return False

    @staticmethod
    def _is_reserved_rpc(content) -> bool:
        rpc_method = content[DATA_PARAMETER][RPC_METHOD_PARAMETER].lower()
        if rpc_method == 'get' or rpc_method == 'set' and RPCRequest._is_device_rpc(content):
            return True

        return False

    @staticmethod
    def _is_device_rpc(content) -> bool:
        return content.get(DEVICE_SECTION_PARAMETER) is not None

    def _fill_rpc_request(self, content):
        if self.rpc_type == RPCType.CONNECTOR:
            self._fill_connector_rpc_request(content)
        elif self.rpc_type == RPCType.DEVICE:
            self._fill_device_rpc_request(content)
        elif self.rpc_type == RPCType.RESERVED:
            self._fill_reserved_rpc_request(content)

    def _fill_connector_rpc_request(self, content):
        (connector_type, rpc_method_name) = content[DATA_PARAMETER][RPC_METHOD_PARAMETER].split('_')
        if connector_type == 'modbus':
            self.method = rpc_method_name
            self.params = content[DATA_PARAMETER][RPC_PARAMS_PARAMETER]
            self._value = content[DATA_PARAMETER][RPC_PARAMS_PARAMETER].get('value')

    def _fill_device_rpc_request(self, content):
        self.method = content[DATA_PARAMETER][RPC_METHOD_PARAMETER]
        self._value = content[DATA_PARAMETER][RPC_PARAMS_PARAMETER]

    def _fill_reserved_rpc_request(self, content):
        self.method = content[DATA_PARAMETER][RPC_METHOD_PARAMETER]

        params = {}

        if self.method == 'set':
            input_params_and_value_list = content[DATA_PARAMETER][RPC_PARAMS_PARAMETER].split(' ')
            if len(input_params_and_value_list) == 1:
                if ";value" in input_params_and_value_list[0]:
                    input_params_and_value_list = input_params_and_value_list[0].split(';value=')
                if len(input_params_and_value_list) == 1:
                    input_params_and_value_list = input_params_and_value_list[0].split(';value')
                input_params_and_value_list[1] = input_params_and_value_list[1].replace(";", '')
            if len(input_params_and_value_list) < 2:
                raise ValueError('Invalid RPC request format. '
                                 'Expected RPC request format: '
                                 'set param_name1=param_value1;param_name2=param_value2;...; value')

            (input_params, input_value) = input_params_and_value_list
            self._value = input_value

        if self.method == 'get':
            input_params = content.get(DATA_PARAMETER, {}).get(RPC_PARAMS_PARAMETER, {})

        for param in input_params.split(';'):
            try:
                (key, value) = param.split('=')
            except ValueError:
                continue

            if key and value:
                params[key] = value if key not in ('functionCode', 'objectsCount', 'address') else int(
                    value)

        self.params = params

    def can_return_response(self):
        return self.id is not None or self.rpc_type == RPCType.CONNECTOR

    def for_existing_device(self):
        return not self.params.get('host') and not self.params.get('port') and not self.params.get('unitId')



#      Copyright 2025. ThingsBoard
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from xknx.telegram.address import GroupAddressType
from xknx.io.const import DEFAULT_MCAST_PORT, DEFAULT_MCAST_GRP
from xknx.io.connection import ConnectionConfig, ConnectionType, SecureConfig

ADDRESS_FORMAT = {
    "SHORT": GroupAddressType.SHORT,
    "LONG": GroupAddressType.LONG,
    "FREE": GroupAddressType.FREE
}

CONNECTION_TYPE = {
    "AUTOMATIC": ConnectionType.AUTOMATIC,
    "ROUTING": ConnectionType.ROUTING,
    "ROUTING_SECURE": ConnectionType.ROUTING_SECURE,
    "TUNNELING": ConnectionType.TUNNELING,
    "TUNNELING_TCP": ConnectionType.TUNNELING_TCP,
    "TUNNELING_TCP_SECURE": ConnectionType.TUNNELING_TCP_SECURE
}


class ClientConfig:
    def __init__(self, config):
        address_format_config = config.get('addressFormat', 'LONG').upper()
        self.address_format = ADDRESS_FORMAT.get(address_format_config, GroupAddressType.LONG)
        self.rate_limit = int(config.get('rateLimit', 0))
        self.multicast_group = config.get('multicastGroup', DEFAULT_MCAST_GRP)
        self.multicast_port = int(config.get('multicastPort', DEFAULT_MCAST_PORT))
        self.state_updater = bool(config.get('stateUpdater', False))
        self.connection_config = self.get_connection_config(config)

    def get_connection_config(self, config):
        security_config = self.get_security_config(config)

        connection_type = config.get('type', 'AUTOMATIC').upper()
        connection_config = ConnectionConfig(connection_type=CONNECTION_TYPE.get(connection_type, "AUTOMATIC"),
                                             individual_address=config.get('individualAddress'),
                                             local_ip=config.get('localIp'),
                                             local_port=int(config.get('localPort', 0)),
                                             gateway_ip=config.get('gatewayIp'),
                                             gateway_port=int(config.get('gatewayPort', DEFAULT_MCAST_PORT)),
                                             multicast_group=config.get('multicastGroup', DEFAULT_MCAST_GRP),
                                             multicast_port=int(config.get('multicastPort', DEFAULT_MCAST_PORT)),
                                             auto_reconnect=bool(config.get('autoReconnect', True)),
                                             auto_reconnect_wait=int(config.get('autoReconnectWait', 3)),
                                             secure_config=security_config)

        return connection_config

    def get_security_config(self, config):
        if config.get('connectionType') in ('AUTOMATIC', 'ROUTING_SECURE', 'TUNNELING_TCP_SECURE'):
            security_config_dict = config.get('security', {})
            security_config = SecureConfig(backbone_key=security_config_dict.get('backboneKey'),
                                           latency_ms=security_config_dict.get('latencyMs'),
                                           user_id=security_config_dict.get('userId'),
                                           device_authentication_password=security_config_dict.get('deviceAuthenticationPassword'),  # noqa
                                           user_password=security_config_dict.get('userPassword'),
                                           knxkeys_file_path=security_config_dict.get('knxkeysFilePath'),
                                           knxkeys_password=security_config_dict.get('knxkeysPassword'))
            return security_config

        return None



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class DeviceInfo:
    def __init__(self, device_info_config, device_details):
        self.device_name = self.__parse_device_name(device_info_config, device_details)
        self.device_type = self.__parse_device_type(device_info_config, device_details)

    def __parse_device_name(self, config, data):
        if config.get('deviceNameExpressionSource', 'expression') == 'expression':
            return self.__parse_device_info(config['deviceNameExpression'], data)

        return config['deviceNameExpression']

    def __parse_device_type(self, config, data):
        if config.get('deviceProfileExpressionSource', 'expression') == 'expression':
            return self.__parse_device_info(config.get('deviceProfileExpression', 'default'), data)

        return config.get('deviceProfileExpression', 'default')

    @staticmethod
    def __parse_device_info(expression, data):
        result_tags = TBUtility.get_values(expression, data.as_dict, get_tag=True)
        result_values = TBUtility.get_values(expression, data.as_dict, expression_instead_none=True)

        result = expression
        for (result_tag, result_value) in zip(result_tags, result_values):
            is_valid_key = "${" in expression and "}" in expression
            result = result.replace('${' + str(result_tag) + '}',
                                    str(result_value)) if is_valid_key else result_tag

        return result



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License"];
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import struct

from thingsboard_gateway.connectors.can.can_converter import CanConverter
from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class BytesCanUplinkConverter(CanConverter):
    def __init__(self, logger):
        self._log = logger

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, configs, can_data):
        device_name = configs.get("deviceName")
        device_type = configs.get("deviceType")

        converted_data = ConvertedData(device_name=device_name, device_type=device_type)

        device_report_strategy = None
        try:
            device_report_strategy = ReportStrategyConfig(configs.get(REPORT_STRATEGY_PARAMETER))
        except ValueError as e:
            self._log.trace("Report strategy config is not specified for device %s: %s", device_name, e)

        for config in configs.get('configs', []):
            try:
                tb_key = config["key"]
                tb_item = "telemetry" if config["is_ts"] else "attributes"

                data_length = config["length"] if config.get("length") is not None and config["length"] != -1 else len(can_data) - config["start"]

                # The 'value' variable is used in eval
                if config["type"][0] == "b":
                    value = bool(can_data[config["start"]])
                elif config["type"][0] == "i" or config["type"][0] == "l":
                    value = int.from_bytes(can_data[config["start"]:config["start"] + data_length],
                                           config["byteorder"],
                                           signed=config["signed"])
                elif config["type"][0] == "f" or config["type"][0] == "d":
                    fmt = ">" + config["type"][0] if config["byteorder"][0] == "b" else "<" + config["type"][0]
                    value = struct.unpack_from(fmt,
                                               bytes(can_data[config["start"]:config["start"] + data_length]))[0]
                elif config["type"][0] == "s":
                    value = can_data[config["start"]:config["start"] + data_length].decode(config["encoding"])
                elif config["type"][0] == "r":
                    value = ""
                    for hex_byte in can_data[config["start"]:config["start"] + data_length]:
                        value += "%02x" % hex_byte
                else:
                    self._log.error("Failed to convert CAN data to TB %s '%s': unknown data type '%s'",
                                    "time series key" if config["is_ts"] else "attribute", tb_key, config["type"])
                    continue

                if config.get("expression", ""):
                    value = eval(config["expression"],
                                                   {"__builtins__": {}} if config["strictEval"] else globals(),
                                                   {"value": value, "can_data": can_data})

                datapoint_key = TBUtility.convert_key_to_datapoint_key(tb_key, device_report_strategy,
                                                                       configs, self._log)
                if tb_item == "attributes":
                    converted_data.add_to_attributes(datapoint_key, value)
                else:
                    telemetry_entry = TelemetryEntry({datapoint_key: value})
                    converted_data.add_to_telemetry(telemetry_entry)
            except Exception as e:
                StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                self._log.error("Failed to convert CAN data to TB %s '%s': %s",
                                "time series key" if config["is_ts"] else "attribute", tb_key, str(e))
                continue

        StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)

        return converted_data



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from logging import getLogger
from simplejson import dumps

from thingsboard_gateway.connectors.modbus.constants import (
    BAUDRATE_PARAMETER,
    BYTE_ORDER_PARAMETER,
    HOST_PARAMETER,
    METHOD_PARAMETER,
    PORT_PARAMETER,
    RETRIES_PARAMETER,
    TIMEOUT_PARAMETER,
    WORD_ORDER_PARAMETER
)
from thingsboard_gateway.gateway.constants import TYPE_PARAMETER


class BackwardCompatibilityAdapter:
    config_files_count = 1
    CONFIG_PATH = None

    def __init__(self, config, config_dir, logger=None):
        if logger:
            self._log = logger
        else:
            self._log = getLogger('BackwardCompatibilityAdapter')

        self.__config = config
        self.__config_dir = config_dir
        BackwardCompatibilityAdapter.CONFIG_PATH = self.__config_dir
        self.__keys = [HOST_PARAMETER, PORT_PARAMETER, TYPE_PARAMETER, METHOD_PARAMETER, TIMEOUT_PARAMETER,
                       BYTE_ORDER_PARAMETER, WORD_ORDER_PARAMETER, RETRIES_PARAMETER, BAUDRATE_PARAMETER]

    @staticmethod
    def __save_json_config_file(config):
        path_to_file = '%rmodbus_new_%r.json' % (BackwardCompatibilityAdapter.CONFIG_PATH,
                                                 BackwardCompatibilityAdapter.config_files_count)
        with open(path_to_file, 'w') as file:
            file.writelines(dumps(config, sort_keys=False, indent='  ', separators=(',', ': ')))
        BackwardCompatibilityAdapter.config_files_count += 1

    def __check_slaves_type_connection(self, config):
        is_tcp_or_udp_connection = False
        is_serial_connection = False

        for slave in config['master']['slaves']:
            if slave[TYPE_PARAMETER] == 'tcp' or slave[TYPE_PARAMETER] == 'udp':
                is_tcp_or_udp_connection = True
            elif slave[TYPE_PARAMETER] == 'serial':
                is_serial_connection = True

        if is_tcp_or_udp_connection and is_serial_connection:
            self._log.warning('It seems that your slaves using different connection type (tcp/udp and serial). '
                              'It is recommended to separate tcp/udp slaves and serial slaves in different connectors '
                              'to avoid problems with reading data.')

    @staticmethod
    def _convert_slave_configuration(slave_config):
        if not slave_config:
            return {}

        values = slave_config.pop('values', {})
        slave_config['values'] = {}
        if len(values):
            for (key, value) in values.items():
                if isinstance(value, list):
                    value = value[0]
                for section_name, section_value in value.items():
                    if slave_config['values'].get(key) is None:
                        slave_config['values'][key] = {section_name: section_value}
                    else:
                        slave_config['values'][key].update({section_name: section_value})

        return slave_config

    def convert(self):
        if not self.__config.get('server'):
            # check if slaves are similar type connection
            self.__check_slaves_type_connection(self.__config)
            converted_slave_config = self._convert_slave_configuration(self.__config.get('slave'))
            self.__config['slave'] = converted_slave_config
            return self.__config

        self._log.warning(
            'You are using old configuration structure for Modbus connector. It will be DEPRECATED in the future '
            'version! New config file "modbus_new.json" was generated in %s folder. Please, use it.', self.CONFIG_PATH)
        self._log.warning('You have to manually connect the new generated config file to tb_gateway.json!')

        slaves = []
        for device in self.__config['server'].get('devices', []):
            slave = {**device}

            for key in self.__keys:
                if not device.get(key):
                    slave[key] = self.__config['server'].get(key)

            slave['pollPeriod'] = slave['timeseriesPollPeriod']

            slaves.append(slave)

        result_dict = {'master': {'slaves': slaves}, 'slave': self.__config.get('slave')}

        # check if slaves are similar type connection
        self.__check_slaves_type_connection(result_dict)

        self.__save_json_config_file(result_dict)

        return result_dict



#      Copyright 2025. ThingsBoard
#  #
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#  #
#          http://www.apache.org/licenses/LICENSE-2.0
#  #
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService


class CollectStatistics:
    def __init__(self, start_stat_type, end_stat_type=None):
        self.start_stat_type = start_stat_type
        self.end_stat_type = end_stat_type

    def __call__(self, func):
        def inner(*args, **kwargs):
            try:
                _, __, data = args
                self.collect(self.start_stat_type, data)
            except ValueError:
                pass

            result = func(*args, **kwargs)
            if result and self.end_stat_type:
                self.collect(self.end_stat_type, result)

            return result

        return inner

    @staticmethod
    def collect(stat_type, data):
        bytes_count = str(data).__sizeof__()
        StatisticsService.add_bytes(stat_type, bytes_count)


class CollectAllReceivedBytesStatistics(CollectStatistics):
    def __call__(self, func):
        def inner(*args, **kwargs):
            try:
                _, data = args
                self.collect(self.start_stat_type, data)
            except ValueError:
                pass

            result = func(*args, **kwargs)
            return result

        return inner


class CollectAllSentTBBytesStatistics(CollectAllReceivedBytesStatistics):
    def __call__(self, func):
        return super().__call__(func)


class CollectRPCReplyStatistics(CollectStatistics):
    def __call__(self, func):
        def inner(*args, **kwargs):
            data = kwargs.get('content')
            if data:
                self.collect(self.start_stat_type, data)

            func(*args, **kwargs)

        return inner


class CollectStorageEventsStatistics(CollectStatistics):
    def __call__(self, func):
        def inner(*args, **kwargs):
            try:
                data = args[1]
                connector_name = args[2]
                if data:
                    StatisticsService.count_connector_message(connector_name, self.start_stat_type)
            except IndexError:
                pass

            func(*args, **kwargs)

        return inner


class CountMessage(CollectStatistics):
    def __call__(self, func):
        def inner(*args, **kwargs):
            StatisticsService.add_count(self.start_stat_type)
            result = func(*args, **kwargs)
            return result

        return inner



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import sqlite3
from sqlite3 import connect, Connection
from threading import Lock
from time import sleep
from typing import Optional


class DatabaseConnector:
    def __init__(self, data_file_path, logger, database_stopped_event):
        self.__log = logger
        self.data_file_path = data_file_path
        self.connection: Optional[Connection] = None
        self.lock = Lock()
        self.database_stopped_event = database_stopped_event
        self.__closed = True

    def connect(self):
        """
        Create database file in path from settings
        """
        try:
            with self.lock:
                self.connection = connect(self.data_file_path, check_same_thread=False)
                self.connection.execute("PRAGMA journal_mode=WAL;")
                self.connection.execute("PRAGMA synchronous=NORMAL;")
                self.connection.execute("PRAGMA cache_size=-20000;")
                self.connection.execute("PRAGMA temp_store=MEMORY;")
                self.connection.execute("PRAGMA journal_size_limit=5000000;")

                self.connection.execute("PRAGMA mmap_size=536870912;")
                self.connection.execute("PRAGMA busy_timeout=15000;")
                self.connection.execute("PRAGMA page_size=4096;")
                self.connection.execute("PRAGMA locking_mode=NORMAL;")
                self.connection.execute("PRAGMA read_uncommitted=ON;")
                self.connection.execute("PRAGMA auto_vacuum=INCREMENTAL;")
                self.connection.execute("PRAGMA foreign_keys=ON;")
                self.connection.row_factory = sqlite3.Row
                self.__closed = False
        except Exception as e:
            self.__log.exception(
                "Failed to connect reading connection to database", exc_info=e
            )

    def connect_on_closed_db(self, database_path):
        try:
            with self.lock:
                self.closed_db_connection = connect(
                    database_path, check_same_thread=False
                )
        except Exception as e:
            self.__log.exception(
                "Failed to connect reading connection to database", exc_info=e
            )

    def commit(self):
        if self.__closed or self.database_stopped_event.is_set():
            return False
        retry_count = 0
        max_retries = 5
        base_sleep_time = 0.05

        while retry_count < max_retries and not self.database_stopped_event.is_set():
            try:
                while (
                    not self.database_stopped_event.is_set()
                    and self.connection is None
                    and not self.__closed
                ):
                    self.__log.debug("Connection is None. Waiting for connection...")
                    sleep(0.1)
                if not self.database_stopped_event.is_set() and not self.__closed:
                    with self.lock:
                        self.connection.commit()
                return True
            except sqlite3.OperationalError:
                sleep_time = base_sleep_time * (2**retry_count)
                self.__log.warning(
                    "Database locked! Retrying in %.2f ms...", sleep_time * 1000
                )
                sleep(sleep_time)
                retry_count += 1
            except Exception as e:
                self.__log.exception("Unexpected error during commit", exc_info=e)
                return False
        return False

    def execute_read(self, *args):
        """
        Execute read query to database
        """
        if self.__closed:
            return None
        try:
            while (
                not self.database_stopped_event.is_set()
                and self.connection is None
                and not self.__closed
            ):
                self.__log.debug("Connection is None. Waiting for connection...")
                sleep(0.1)
            if (
                not self.database_stopped_event.is_set()
                and not self.__closed
                and self.connection is not None
            ):
                with self.lock:
                    cursor = self.connection.cursor()
                    result = cursor.execute(*args)
                    return result
        except sqlite3.ProgrammingError as e:
            self.__log.debug("Failed to execute read query in database", exc_info=e)
        except sqlite3.OperationalError as e:
            self.__log.debug("Failed to execute read query in database", exc_info=e)
        except Exception as e:
            self.__log.exception("Failed to execute read query in database", exc_info=e)

    def execute_write(self, *args):
        """
        Execute write query to database
        """
        if self.__closed:
            return None
        tries = 4
        current_try = 0
        while current_try < tries:
            try:
                while (
                    not self.database_stopped_event.is_set()
                    and self.connection is None
                    and not self.__closed
                ):
                    self.__log.debug("Connection is None. Waiting for connection...")
                    sleep(0.1)
                if (
                    not self.database_stopped_event.is_set()
                    and not self.__closed
                    and self.connection is not None
                ):
                    with self.lock:
                        cursor = self.connection.cursor()
                        result = cursor.execute(*args)
                        return result
            except sqlite3.OperationalError:
                current_try += 1
                sleep(0.05 * current_try)
            except sqlite3.ProgrammingError as e:
                self.__log.debug(
                    "Failed to execute write query in database", exc_info=e
                )
                current_try += 1
                sleep(0.05 * current_try)
            except Exception as e:
                self.__log.exception(
                    "Failed to execute write query in database", exc_info=e
                )
                current_try += 1
                sleep(0.05 * current_try)

    def execute_many_write(self, *args):
        """
        Execute write query with many records to database
        """
        if self.__closed:
            return None
        tries = 4
        current_try = 0
        while current_try < tries:
            try:
                while (
                    not self.database_stopped_event.is_set() and self.connection is None
                ):
                    self.__log.debug("Connection is None. Waiting for connection...")
                    sleep(0.1)
                if (
                    not self.database_stopped_event.is_set()
                    and not self.__closed
                    and self.connection is not None
                ):
                    with self.lock:
                        cursor = self.connection.cursor()
                        cursor.execute("BEGIN TRANSACTION;")
                        return cursor.executemany(*args)
            except sqlite3.OperationalError:
                current_try += 1
                sleep(0.05 * current_try)
            except sqlite3.ProgrammingError as e:
                self.__log.debug(
                    "Failed to execute write query in database", exc_info=e
                )
                current_try += 1
                sleep(0.05 * current_try)
            except Exception as e:
                self.__log.exception(
                    "Failed to execute write query in database", exc_info=e
                )
                current_try += 1
                sleep(0.05 * current_try)

    def rollback(self):
        """
        Rollback changes after exception
        """
        if self.__closed:
            return
        self.__log.info("Rollback transaction")
        self.connection.rollback()

    def interrupt(self):
        """
        Interrupts database connection
        """
        if not self.__closed:
            self.__log.debug("Interrupting connection to database")
            self.connection.interrupt()

    @property
    def closed(self):
        return self.__closed

    def close(self):
        """
        Closes database file
        """
        if not self.__closed:
            self.__closed = True
            if self.connection.in_transaction:
                self.connection.interrupt()
            try:
                if self.connection is not None:
                    with self.lock:
                        self.__log.debug("Closing connection to database")
                        self.connection.close()
                        self.connection = None
            except Exception as e:
                self.__log.exception(
                    "Failed to close connection to database", exc_info=e
                )

    def get_cursor(self):
        try:
            return self.connection.cursor()

        except Exception as e:
            self.__log.exception("Failed to get cursor", exc_info=e)

    def update_logger(self, logger):
        self.__log = logger



#      Copyright 2025. ThingsBoard
#  #
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#  #
#          http://www.apache.org/licenses/LICENSE-2.0
#  #
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import datetime
import subprocess
from threading import Thread, RLock, Event
from time import monotonic, sleep
from platform import system as platform_system

import simplejson

from thingsboard_gateway.gateway.statistics.configs import ONCE_SEND_STATISTICS_CONFIG, SERVICE_STATS_CONFIG, \
    MACHINE_STATS_CONFIG


class StatisticsService(Thread):
    ENABLED = False
    ENABLED_CUSTOM = False
    STATISTICS_STORAGE = {
        'receivedBytesFromDevices': 0,
        'convertedBytesFromDevice': 0,
        'allReceivedBytesFromTB': 0,
        'allBytesSentToTB': 0,
        'allBytesSentToDevices': 0,
        'eventsAdded': 0,
        'eventsProcessed': 0,
        'msgsSentToPlatform': 0,
        'msgsReceivedFromPlatform': 0,
        'storageMsgPulled': 0,
        'platformMsgPushed': 0,
        'platformAttrProduced': 0,
        'platformTsProduced': 0,
    }

    # This is a dictionary that stores the statistics for each connector
    # The key is the connector name and the value is another dictionary
    # The key of the inner dictionary is the statistic name and the value is the statistic value
    # Example:
    # {
    #     "connector_name": {
    #         "statistic_name": statistic_value
    #         ...
    #     }
    # }
    CONNECTOR_STATISTICS_STORAGE = {}
    __LOCK = RLock()

    def __init__(self, statistics_configuration, gateway, log, config_path=None):
        stats_send_period_in_seconds = statistics_configuration['statsSendPeriodInSeconds']
        self._custom_stats_send_period_in_seconds = statistics_configuration.get('customStatsSendPeriodInSeconds', 300)
        super().__init__()
        self.name = 'Statistics Thread'
        self.daemon = True
        self._stopped = Event()

        self._config_path = config_path
        self._stats_send_period_in_seconds = stats_send_period_in_seconds if stats_send_period_in_seconds >= 10 else 10
        self._gateway = gateway
        self._log = log
        self._custom_command_config = self._load_config()
        self.__install_required_tools()
        self._last_service_poll = 0
        self._last_custom_command_poll = 0
        self._last_streams_statistics_clear_time = datetime.datetime.now()

        self.start()

    def stop(self):
        self._stopped.set()

    def _load_config(self):
        config = []

        # load custom statistics config
        if self._config_path:
            with open(self._config_path, 'r') as file:
                config = simplejson.load(file)

        return config

    @classmethod
    def enable_statistics(cls):
        with cls.__LOCK:
            cls.ENABLED = True

    @classmethod
    def disable_statistics(cls):
        with cls.__LOCK:
            cls.ENABLED = False

    @classmethod
    def enable_custom_statistics(cls):
        with cls.__LOCK:
            cls.ENABLED_CUSTOM = True

    @classmethod
    def disable_custom_statistics(cls):
        with cls.__LOCK:
            cls.ENABLED_CUSTOM = False

    @classmethod
    def add_bytes(cls, stat_type, bytes_count, stat_parameter_name=None, statistics_type='STATISTICS_STORAGE'):
        if StatisticsService.ENABLED:
            cls.add_count(stat_type, bytes_count, stat_parameter_name, statistics_type)

    @classmethod
    def add_count(cls, stat_key, count=1, stat_parameter_name=None, statistics_type='STATISTICS_STORAGE'):
        if StatisticsService.ENABLED:
            if statistics_type == 'CONNECTOR_STATISTICS_STORAGE':
                try:
                    if cls.CONNECTOR_STATISTICS_STORAGE.get(stat_key) is None:
                        cls.CONNECTOR_STATISTICS_STORAGE[stat_key] = {}

                    cls.CONNECTOR_STATISTICS_STORAGE[stat_key][stat_parameter_name] += count
                except KeyError:
                    cls.CONNECTOR_STATISTICS_STORAGE[stat_key].update({stat_parameter_name: count})
            else:
                cls.STATISTICS_STORAGE[stat_key] += count

    @classmethod
    def clear_statistics(cls):
        with cls.__LOCK:
            for key in cls.STATISTICS_STORAGE:
                cls.STATISTICS_STORAGE[key] = 0
            cls.CONNECTOR_STATISTICS_STORAGE = {}

    @staticmethod
    def count_connector_message(connector_name, stat_parameter_name, count=1):
        if StatisticsService.ENABLED:
            StatisticsService.add_count(connector_name, stat_parameter_name=stat_parameter_name,
                                        statistics_type='CONNECTOR_STATISTICS_STORAGE', count=count)

    @staticmethod
    def count_connector_bytes(connector_name, msg, stat_parameter_name):
        if StatisticsService.ENABLED:
            bytes_count = str(msg).__sizeof__()
            StatisticsService.add_bytes(connector_name, bytes_count, stat_parameter_name,
                                        statistics_type='CONNECTOR_STATISTICS_STORAGE')

    def __install_required_tools(self):
        if self._custom_command_config:
            for attribute in self._custom_command_config:
                installation_command = attribute.get('installCmd')
                if installation_command:
                    try:
                        if platform_system() == 'Windows':
                            subprocess.run(installation_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                           encoding='utf-8', timeout=attribute['timeout'])
                        else:
                            subprocess.run(['/bin/sh', '-c', installation_command], stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE, encoding='utf-8', timeout=attribute['timeout'])
                    except Exception as e:
                        self._log.error("Error while executing installation command '%s': %s", installation_command, e)

    def __collect_custom_command_statistics(self):
        message = {}
        for attribute in self._custom_command_config:
            try:
                if attribute.get('command') is not None:
                    if platform_system() == 'Windows':
                        process = subprocess.run(attribute['command'], stdout=subprocess.PIPE,
                                                 stderr=subprocess.PIPE,
                                                 encoding='utf-8', timeout=attribute['timeout'])
                    else:
                        process = subprocess.run(['/bin/sh', '-c', attribute['command']],
                                                 stdout=subprocess.PIPE,
                                                 stderr=subprocess.PIPE,
                                                 encoding='utf-8', timeout=attribute['timeout'])

                    value = process.stdout
                else:
                    value = attribute['function'](self._gateway)
                if not value:
                    continue
                try:
                    value = float(value)
                except ValueError:
                    pass
            except Exception as e:
                self._log.warning("Statistic parameter %s raise the exception: %s",
                                  attribute['attributeOnGateway'], e)
                continue
            message[attribute['attributeOnGateway']] = value

        return message

    def __collect_statistics_from_config(self, statistics_config):
        message = {}

        for attribute in statistics_config:
            try:
                message[attribute['attributeOnGateway']] = attribute['function'](self._gateway)
            except Exception as e:
                self._log.warning("Statistic parameter %s raise the exception: %s",
                                  attribute['attributeOnGateway'], e)
                continue

        return message

    def __collect_service_statistics(self):
        message = {}

        for attribute in SERVICE_STATS_CONFIG:
            try:
                message[attribute['attributeOnGateway']] = attribute['function'](self._gateway)
            except Exception as e:
                self._log.warning("Statistic parameter %s raise the exception: %s",
                                  attribute['attributeOnGateway'], e)
                continue

        return message

    def __send_custom_command_statistics(self):
        if self.ENABLED_CUSTOM:
            custom_command_statistics_message = self.__collect_custom_command_statistics()
            if custom_command_statistics_message:
                self._gateway.send_telemetry(custom_command_statistics_message)

    def __send_statistics(self):
        statistics_message = {'machineStats': self.__collect_statistics_from_config(MACHINE_STATS_CONFIG),
                              'serviceStats': self.__collect_service_statistics(),
                              'connectorsStats': self.CONNECTOR_STATISTICS_STORAGE}
        self._log.info('Collected regular statistics: %s', statistics_message)
        if StatisticsService.ENABLED:
            self._gateway.send_telemetry(statistics_message)
        else:
            self._log.debug('Statistics are disabled. Not sending collected statistics.')
        self.clear_statistics()

    def __send_general_machine_state(self):
        message = self.__collect_statistics_from_config(ONCE_SEND_STATISTICS_CONFIG)
        self._gateway.send_attributes(message)

    def __once_send_statistics(self):
        self.__send_general_machine_state()

    def run(self) -> None:
        try:
            self.__once_send_statistics()
        except Exception as e:
            self._log.error("Error while sending first statistics information: %s", e)

        while not self._stopped.is_set() and (StatisticsService.ENABLED or StatisticsService.ENABLED_CUSTOM):
            try:
                if not StatisticsService.ENABLED and not StatisticsService.ENABLED_CUSTOM:
                    self._log.debug("Statistics service is disabled. Stopping the thread.")
                    break
                cur_monotonic = int(monotonic())

                next_service_poll = (self._last_service_poll + self._stats_send_period_in_seconds) - cur_monotonic
                next_custom_command_poll = (self._last_custom_command_poll +
                                            self._custom_stats_send_period_in_seconds) - cur_monotonic

                wait_time = max(0, min(next_service_poll, next_custom_command_poll))

                if wait_time > 0:
                    sleep(wait_time)

                if not StatisticsService.ENABLED and not StatisticsService.ENABLED_CUSTOM:
                    continue

                cur_monotonic = int(monotonic())
                if (cur_monotonic - self._last_service_poll >= self._stats_send_period_in_seconds or
                        self._last_service_poll == 0):
                    self._last_service_poll = cur_monotonic
                    self.__send_statistics()

                if (cur_monotonic - self._last_custom_command_poll >= self._custom_stats_send_period_in_seconds or
                        self._last_custom_command_poll == 0):
                    self._last_custom_command_poll = cur_monotonic
                    self.__send_custom_command_statistics()

            except Exception as e:
                self._log.error("Error in statistics thread: %s", e)
                self._stopped.wait(5)
        self.stop()



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import asyncio
from asyncio import CancelledError, Queue as AsyncQueue, QueueEmpty
from queue import Queue, Empty
from threading import Thread
from random import choice
from string import ascii_lowercase
from time import monotonic, sleep

from packaging import version

from thingsboard_gateway.connectors.modbus.constants import ADDRESS_PARAMETER, TAG_PARAMETER, \
    FUNCTION_CODE_PARAMETER
from thingsboard_gateway.connectors.modbus.entities.rpc_request import RPCRequest, RPCType
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.gateway.constants import STATISTIC_MESSAGE_RECEIVED_PARAMETER, \
    STATISTIC_MESSAGE_SENT_PARAMETER, CONNECTOR_PARAMETER, DEVICE_SECTION_PARAMETER, DATA_PARAMETER, \
    RPC_METHOD_PARAMETER, RPC_PARAMS_PARAMETER
from thingsboard_gateway.tb_utility.tb_logger import init_logger

# Try import Pymodbus library or install it and import
installation_required = False
required_version = '3.9.2'
force_install = False

try:
    from pymodbus import __version__ as pymodbus_version

    if version.parse(pymodbus_version) != version.parse(required_version):
        installation_required = True
        force_install = True

    from serial import SerialException  # noqa
    from serial_asyncio import create_serial_connection  # noqa

except ImportError:
    installation_required = True

if installation_required:
    print("Modbus library not found - installing...")
    TBUtility.install_package("pymodbus", required_version, force_install=force_install)
    TBUtility.install_package('pyserial')
    TBUtility.install_package('pyserial-asyncio')

from thingsboard_gateway.connectors.modbus.entities.master import Master  # noqa: E402
from thingsboard_gateway.connectors.modbus.server import Server  # noqa: E402
from thingsboard_gateway.connectors.modbus.slave import Slave  # noqa: E402
from thingsboard_gateway.connectors.modbus.entities.bytes_downlink_converter_config import \
    BytesDownlinkConverterConfig  # noqa: E402
from thingsboard_gateway.connectors.modbus.backward_compatibility_adapter import BackwardCompatibilityAdapter  # noqa

from pymodbus.exceptions import ConnectionException  # noqa: E402
from pymodbus.pdu import ModbusPDU  # noqa
from pymodbus.pdu.bit_message import WriteMultipleCoilsResponse, WriteSingleCoilResponse  # noqa: E402
from pymodbus.constants import Endian  # noqa: E402
from pymodbus.pdu import ExceptionResponse  # noqa: E402
from pymodbus.pdu.register_message import WriteMultipleRegistersResponse, WriteSingleRegisterResponse  # noqa: E402


class AsyncModbusConnector(Connector, Thread):
    def __init__(self, gateway, config, connector_type):
        self.statistics = {STATISTIC_MESSAGE_RECEIVED_PARAMETER: 0,
                           STATISTIC_MESSAGE_SENT_PARAMETER: 0}
        super().__init__()
        self.__gateway = gateway
        self._connector_type = connector_type
        self.name = config.get("name", 'Modbus Connector ' + ''.join(choice(ascii_lowercase) for _ in range(5)))
        self.__log = init_logger(self.__gateway, config.get('name', self.name),
                                 config.get('logLevel', 'INFO'),
                                 enable_remote_logging=config.get('enableRemoteLogging', False),
                                 is_connector_logger=True)
        self.converter_log = init_logger(self.__gateway, self.name + '_converter',
                                         config.get('logLevel', 'INFO'),
                                         enable_remote_logging=config.get('enableRemoteLogging', False),
                                         is_converter_logger=True, attr_name=self.name)
        self.__backward_compatibility_adapter = BackwardCompatibilityAdapter(config, gateway.get_config_path(),
                                                                             logger=self.__log)
        self.__config = self.__backward_compatibility_adapter.convert()
        self.__log.info('Starting Modbus Connector...')
        self.__id = self.__config.get('id')
        self.__connected = False
        self.__stopped = False
        self.daemon = True

        try:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        except RuntimeError:
            self.loop = asyncio.get_event_loop()

        self.process_device_requests = AsyncQueue(100000)
        self.__data_to_convert = Queue(-1)
        self.__data_to_save = Queue(-1)

        self.__server = None

        self._master_connections = {}

        self.__slaves = []

    def close(self):
        self.__stopped = True
        self.__connected = False

        self.__log.debug('Stopping %s...', self.get_name())

        for slave in self.__slaves:
            slave.close(self.loop)

        if self.__server:
            self.__server.stop()

        asyncio.run_coroutine_threadsafe(self.__cancel_all_tasks(), self.loop)

        self.__check_is_alive()

        self.__log.info('%s has been stopped', self.get_name())
        self.__log.stop()

    def __check_is_alive(self):
        start_time = monotonic()

        while self.is_alive():
            if monotonic() - start_time > 10:
                self.__log.error("Failed to stop connector %s", self.get_name())
                break
            sleep(.1)

    async def __cancel_all_tasks(self):
        await asyncio.sleep(5)
        for task in asyncio.all_tasks(self.loop):
            task.cancel()

    def open(self):
        self.start()

    def run(self):
        self.__connected = True

        try:
            self.loop.run_until_complete(self.__run())
        except CancelledError as e:
            self.__log.debug('Task was cancelled due to connector stop: %s', e.__str__())
        except Exception as e:
            self.__log.exception(e)

    async def __run(self):
        # check if Gateway as a server need to be started
        if Server.is_runnable(self.__config):
            self.__run_server()

        await self.__add_slaves(self.__config.get('master', {'slaves': []}).get('slaves', []))

        await asyncio.gather(
            self.__process_requests(),
            self.__convert_data(),
            self.__save_data()
        )

    def __run_server(self):
        try:
            self.__log.info("Starting Modbus server")
            self.__server = Server(self.__config['slave'], self.__log)
            self.__server.start()
            self.__add_slave(self.__server.get_slave_config_format())
            self.__log.info("Modbus server started")
        except Exception as e:
            self.__log.exception('Failed to start Modbus server: %s', e)
            self.__connected = False

    def __get_master(self, slave: Slave):
        """
        Method check if connection to master already exists and return it
        or create new connection and return it

        Newly created master connection structure will look like:
        self._master_connections = {
            '127.0.0.1:5021': AsyncClient object (TCP/UDP),
            '/dev/ttyUSB0': AsyncClient object (Serial)
        }
        """
        if not slave.host:
            master_connection_name = slave.port
        else:
            master_connection_name = slave.host + ':' + str(slave.port)
        if master_connection_name not in self._master_connections:
            master_connection = Master.configure_master(slave)
            master = Master(slave.type, master_connection)
            self._master_connections[master_connection_name] = master

        return self._master_connections[master_connection_name]

    def __add_slave(self, slave_config):
        slave = Slave(self, self.__log, slave_config)
        master = self.__get_master(slave)
        slave.master = master

        self.__slaves.append(slave)
        slave.start()

    async def __add_slaves(self, slaves_config):
        for slave_config in slaves_config:
            try:
                self.__add_slave(slave_config)
            except Exception as e:
                self.__log.exception('Failed to add slave: %s', e)

        self.__log.debug('Added %d slaves', len(self.__slaves))

    @classmethod
    def callback(cls, slave: Slave, queue: Queue):
        queue.put_nowait(slave)

    async def __process_requests(self):
        while not self.__stopped:
            try:
                slave = self.process_device_requests.get_nowait()

                if slave.type == 'serial':
                    await self.__poll_device(slave)
                else:
                    self.loop.create_task(self.__poll_device(slave))
            except QueueEmpty:
                await asyncio.sleep(0.1)
                continue
            except Exception as e:
                self.__log.exception('Failed to poll device: %s', e)

    async def __poll_device(self, slave: Slave):
        self.__log.debug("Polling %s slave", slave)

        # check if device have attributes or telemetry to poll
        if slave.uplink_converter_config.attributes or slave.uplink_converter_config.telemetry:
            try:
                connected_to_master = await slave.connect()

                if connected_to_master:
                    self.__manage_device_connectivity_to_platform(slave)

                if connected_to_master:
                    slave_data = await self.__read_slave_data(slave)
                    self.__data_to_convert.put_nowait((slave, slave_data))
                else:
                    self.__log.error('Device %s is not connected, cannot connect to server, skipping reading...', slave)
                    self.__delete_device_from_platform(slave)
            except ConnectionException:
                self.__delete_device_from_platform(slave)
                await asyncio.sleep(5)
                self.__log.error('Failed to connect to device %s', slave)
            except asyncio.exceptions.TimeoutError:
                self.__log.error('Timeout error for device %s', slave)
                await slave.disconnect()
            except Exception as e:
                self.__delete_device_from_platform(slave)
                self.__log.error('Failed to poll %s device: %s', slave, e)
                self.__log.debug('Exception:', exc_info=e)
        else:
            self.__log.debug('Config is empty. Nothing to read, for device %s', slave)

    async def __read_slave_data(self, slave: Slave):
        result = {
            'telemetry': {},
            'attributes': {}
        }

        for config_section in ('attributes', 'telemetry'):
            for config in getattr(slave.uplink_converter_config, config_section):
                try:
                    response = await slave.read(config['functionCode'], config['address'], config['objectsCount'])
                except asyncio.exceptions.TimeoutError:
                    self.__log.error("Timeout error for device %s function code %s address %s, it may be caused by wrong data in server register.",  # noqa
                                     slave.device_name, config['functionCode'], config[ADDRESS_PARAMETER])
                    continue

                result[config_section][config['tag']] = response

        return result

    def __manage_device_connectivity_to_platform(self, slave: Slave):
        if slave.master.connected() and slave.device_name not in self.__gateway.get_devices():
            self.__add_device_to_platform(slave)

    def __delete_device_from_platform(self, slave: Slave):
        if slave.device_name in self.__gateway.get_devices():
            self.__gateway.del_device(slave.device_name)
            slave.last_connect_time = 0

    def __add_device_to_platform(self, slave: Slave):
        """
        Add device to platform
        """

        if slave.master.connected():
            device_connected = self.__gateway.add_device(slave.device_name,
                                                         {CONNECTOR_PARAMETER: self},
                                                         device_type=slave.device_type)

            slave.last_connect_time = monotonic() if device_connected else 0

    async def __convert_data(self):
        while not self.__stopped:
            try:
                batch_to_convert = {}
                if not self.__data_to_convert.empty():
                    batch_forming_start = monotonic()
                    while not self.__stopped and monotonic() - batch_forming_start < 0.1:
                        try:
                            slave, data = self.__data_to_convert.get_nowait()
                        except Empty:
                            break

                        batch_key = (slave.device_name, slave.uplink_converter)

                        if batch_key not in batch_to_convert:
                            batch_to_convert[batch_key] = []

                        batch_to_convert[batch_key].append(data)

                    for (device_name, uplink_converter), data in batch_to_convert.items():
                        converted_data: ConvertedData = uplink_converter.convert({}, data)
                        self.__log.trace("Converted data: %r", converted_data)
                        if len(converted_data['attributes']) or len(converted_data['telemetry']):
                            self.__data_to_save.put_nowait(converted_data)
                else:
                    await asyncio.sleep(.1)
            except Exception as e:
                self.__log.error('Exception in conversion data loop: %s', e)

    async def __save_data(self):
        while not self.__stopped:
            if not self.__data_to_save.empty():
                try:
                    converted_data = self.__data_to_save.get_nowait()
                    StatisticsService.count_connector_message(self.get_name(), stat_parameter_name='storageMsgPushed')
                    self.__gateway.send_to_storage(self.get_name(), self.get_id(), converted_data)
                    self.statistics[STATISTIC_MESSAGE_SENT_PARAMETER] += 1
                except Empty:
                    await asyncio.sleep(.001)
            else:
                await asyncio.sleep(.001)

    def get_device_shared_attributes_keys(self, device_name):
        device = self.__get_device_by_name(device_name)
        if device is not None:
            return device.shared_attributes_keys
        return []

    def on_attributes_update(self, content):
        self.__log.debug('Got attributes update: %s', content)

        try:
            device = self.__get_device_by_name(content[DEVICE_SECTION_PARAMETER])
            if device is None:
                self.__log.error('Device %s not found in connector %s',
                                 content[DEVICE_SECTION_PARAMETER],
                                 self.get_name())
                return

            for attribute_updates_command_config in device.attributes_updates_config:
                for attribute_updated in content[DATA_PARAMETER]:
                    if attribute_updates_command_config[TAG_PARAMETER] == attribute_updated:
                        to_process = {
                            DEVICE_SECTION_PARAMETER: content[DEVICE_SECTION_PARAMETER],
                            DATA_PARAMETER: {
                                RPC_METHOD_PARAMETER: attribute_updated,
                                RPC_PARAMS_PARAMETER: content[DATA_PARAMETER][attribute_updated]
                            }
                        }

                        self.__create_task(self.__process_attribute_request,
                                           (device, to_process,
                                            attribute_updates_command_config),
                                           {})
        except Exception as e:
            self.__log.exception('Failed to update attributes: %s', e)

    async def __process_attribute_request(self, device: Slave, data, config):
        if config is not None:
            converted_data = None

            config = BytesDownlinkConverterConfig(
                device_name=device.device_name,
                byte_order=device.byte_order,
                word_order=device.word_order,
                repack=device.repack,
                objects_count=config.get("objectsCount", config.get("registersCount", config.get("registerCount", 1))),
                function_code=config.get(FUNCTION_CODE_PARAMETER),
                lower_type=config.get("type", config.get("tag", "error")),
                address=config.get(ADDRESS_PARAMETER)
            )

            if config.function_code in (5, 6):
                converted_data = device.downlink_converter.convert(config, data)

                try:
                    converted_data = converted_data[0]
                except IndexError and TypeError:
                    pass
            elif config.function_code in (15, 16):
                converted_data = device.downlink_converter.convert(config, data)

            if converted_data is not None:
                try:
                    connected = await device.connect()
                    if connected:
                        await device.write(config.function_code, config.address, converted_data)
                    else:
                        self.__log.error(
                            'Failed to process attribute update: Socket is closed, connection is lost, for device %s',
                            device)
                except Exception as e:
                    self.__log.exception('Failed to process attribute update: ', e)

    def server_side_rpc_handler(self, content):
        self.__log.info('Received server side rpc request: %r', content)

        try:
            rpc_request = RPCRequest(content)

            response = None
            if rpc_request.rpc_type == RPCType.CONNECTOR:
                response = self.__process_connector_rpc_request(rpc_request)
            elif rpc_request.rpc_type == RPCType.RESERVED:
                response = self.__process_reserved_rpc_request(rpc_request)
            elif rpc_request.rpc_type == RPCType.DEVICE:
                response = self.__process_device_rpc_request(rpc_request)

            return response
        except Exception as e:
            self.__log.error('Failed to process server side rpc request: %s', e)
            return {'error': '%r' % e, 'success': False}

    def __process_connector_rpc_request(self, rpc_request: RPCRequest):
        self.__log.debug("Received RPC to connector: %r", rpc_request)

        if rpc_request.for_existing_device():
            results = []
            for device in self.__slaves:
                rpc_request.device_name = device.device_name
                result = {}
                self.__create_task(self.__process_rpc_request,
                                   (device, rpc_request.params, rpc_request),
                                   {'with_response': True, 'result': result},
                                   timeout=rpc_request.timeout)
                results.append(result['response'])

            return results
        else:
            slave = Slave(self, self.__log, rpc_request.params)
            master = self.__get_master(slave)
            slave.master = master
            rpc_request.device_name = slave.device_name

            connected_to_master_task = self.loop.create_task(slave.connect())
            connect_started = monotonic()
            while not connected_to_master_task.done() and not self.__stopped and \
                    monotonic() - connect_started < rpc_request.timeout:
                sleep(.02)

            if connected_to_master_task.result():
                result = {}
                self.__create_task(self.__process_rpc_request,
                                   (slave, rpc_request.params, rpc_request),
                                   {'with_response': True, 'result': result},
                                   timeout=rpc_request.timeout)
                return result['response']
            else:
                self.__log.error('Failed to connect to device %s', slave.device_name)
                return {'error': 'Failed to connect to device %s' % slave.device_name, 'success': False}

    def __process_reserved_rpc_request(self, rpc_request: RPCRequest):
        device = self.__get_device_by_name(rpc_request.device_name)
        if device is None:
            self.__log.error('Device %s not found in connector %s', rpc_request.device_name, self.get_name())
            return {'error': 'Device %s not found' % rpc_request.device_name, 'success': False}

        result = {}
        self.__create_task(self.__process_rpc_request,
                           (device, rpc_request.params, rpc_request),
                           {'result': result},
                           timeout=rpc_request.timeout)

        self.__log.debug("Result: %r", result)

        return result['response']

    def __process_device_rpc_request(self, rpc_request: RPCRequest):
        device = self.__get_device_by_name(rpc_request.device_name)
        if device is None:
            self.__log.error('Device %s not found in connector %s', rpc_request.device_name, self.get_name())
            return {'error': 'Device %s not found' % rpc_request.device_name, 'success': False}

        device_rpc_config = device.get_device_rpc_config(rpc_request.method)
        if device_rpc_config is None:
            self.__log.error("Received rpc request, but method %s not found in config for %s.",
                             rpc_request.method,
                             self.get_name())
            self.__gateway.send_rpc_reply(rpc_request.device_name,
                                          rpc_request.id,
                                          {rpc_request.method: "METHOD NOT FOUND!"})
            return

        result = {}
        self.__create_task(self.__process_rpc_request,
                           (device, device_rpc_config, rpc_request),
                           {'result': result},
                           timeout=rpc_request.timeout)

        self.__log.debug("Result: %r", result)

        return result['response']

    def __create_task(self, func, args, kwargs, timeout=5.0):
        task = self.loop.create_task(func(*args, **kwargs))
        task_started = monotonic()

        while not task.done() and not self.__stopped and monotonic() - task_started < timeout:
            sleep(.02)

    async def __process_rpc_request(self, device: Slave, config, data, with_response=False, result={}):
        try:
            if config is not None:
                if config['functionCode'] in (5, 6, 15, 16):
                    response = await self.__write_rpc_data(device, config, data)
                elif config['functionCode'] in (1, 2, 3, 4):
                    response = await self.__read_rpc_data(device, config)
                else:
                    response = 'Unsupported function code in RPC request.'

                if data.can_return_response():
                    result['response'] = self.__send_rpc_response(data, response, with_response)
        except Exception as e:
            self.__log.error('Failed to process rpc request: %r', e)
            result['response'] = {'error': e.__repr__()}

    async def __read_rpc_data(self, device: Slave, config):
        response = None

        try:
            connected = await device.connect()
            if connected:
                response = await device.read(config['functionCode'], config['address'], config['objectsCount'])
        except Exception as e:
            self.__log.error('Failed to process rpc request: %s', e)
            response['response'] = {'error': e.__repr__()}

        if isinstance(response, ModbusPDU):
            endian_order = Endian.BIG if device.byte_order.upper() == "BIG" else Endian.LITTLE
            word_endian_order = Endian.BIG if device.word_order.upper() == "BIG" else Endian.LITTLE
            response = device.uplink_converter.decode_data(response, config, endian_order, word_endian_order)

        return response

    async def __write_rpc_data(self, device, config, data):
        response = None

        config = BytesDownlinkConverterConfig(
            device_name=device.device_name,
            byte_order=device.byte_order,
            word_order=device.word_order,
            repack=device.repack,
            objects_count=config.get("objectsCount", config.get("registersCount", config.get("registerCount", 1))),
            function_code=config.get(FUNCTION_CODE_PARAMETER),
            lower_type=config.get("type", config.get("tag", "error")),
            address=config.get(ADDRESS_PARAMETER)
        )

        converted_data = device.downlink_converter.convert(config, data.value)

        if config.function_code in (5, 6):
            try:
                converted_data = converted_data[0]
            except IndexError and TypeError:
                pass

        if converted_data is not None:
            try:
                connected = await device.connect()
                if connected:
                    response = await device.write(config.function_code, config.address, converted_data)
            except Exception as e:
                self.__log.error('Failed to process rpc request for device: %s', device, exc_info=e)
                response = e

        if isinstance(response, (WriteMultipleRegistersResponse,
                                 WriteMultipleCoilsResponse,
                                 WriteSingleCoilResponse,
                                 WriteSingleRegisterResponse)):
            self.__log.debug("Write %r", str(response))
            response = {"success": True}

        return response

    def __send_rpc_response(self, content, response, with_response=False):
        method = content.method
        if isinstance(response, Exception) or isinstance(response, ExceptionResponse):
            if not with_response:
                self.__gateway.send_rpc_reply(device=content.device_name,
                                              req_id=content.id,
                                              content={
                                                  method: response.__repr__()
                                              },
                                              success_sent=False)
            else:
                return {
                    'device': content.device_name,
                    'req_id': content.id,
                    'content': {
                        method: str(response)
                    },
                    'success_sent': False
                }
        else:
            if not with_response:
                self.__gateway.send_rpc_reply(device=content.device_name,
                                              req_id=content.id,
                                              content={'result': response})
            else:
                response = {'device': content.device_name, 'content': response}

                if content.id is not None:
                    response['req_id'] = content.id

                return response

    def __get_device_by_name(self, device_name) -> Slave:
        devices = tuple(filter(lambda slave: slave.device_name == device_name, self.__slaves))

        if len(devices) > 0:
            return devices[0]

        return None

    @property
    def connector_type(self):
        return self._connector_type

    def get_id(self):
        return self.__id

    def get_name(self):
        return self.name

    def get_type(self):
        return self._connector_type

    def get_config(self):
        return self.__config

    def is_connected(self):
        return self.__connected

    def is_stopped(self):
        return self.__stopped



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from asyncio import Lock, sleep
from re import escape, match, fullmatch, compile
from time import monotonic

from bacpypes3.primitivedata import ObjectIdentifier

from thingsboard_gateway.connectors.bacnet.bacnet_uplink_converter import AsyncBACnetUplinkConverter
from thingsboard_gateway.connectors.bacnet.entities.bacnet_device_details import BACnetDeviceDetails
from thingsboard_gateway.connectors.bacnet.entities.device_info import DeviceInfo
from thingsboard_gateway.connectors.bacnet.entities.device_object_config import DeviceObjectConfig
from thingsboard_gateway.connectors.bacnet.entities.uplink_converter_config import UplinkConverterConfig
from thingsboard_gateway.gateway.constants import UPLINK_PREFIX, CONVERTER_PARAMETER
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class Device:
    ALLOWED_WITH_ROUTERS_ADDRESSES_PATTERN = compile(r'^([^:,@]+)(:[^:,@]+)?(:(:47808))?$')
    ALLOWED_LOCAL_ADDRESSES_PATTERN_WITH_DEFAULT_PORT = compile(r'^(0:[^:,@]+|[^\d:,@][^:,@]*:47808|[^:,@]+)$')
    ALLOWED_LOCAL_ADDRESSES_PATTERN_WITH_CUSTOM_PORT = (
        r'^((:(0:)?)[^:,@]+|[^\d:,@][^:,@]*|[^:,@]+)'
        r'\:{expectedPort}'
        r'(@[^:,@]+(:\d+)?)?$'
    )
    ANY_LOCAL_ADDRESS_WITH_PORT_PATTERN = compile(r"\*:\d+")
    ROUTER_INFO_PARAMS = [
        '${routerName}',
        '${routerId}',
        '${routerVendorId}',
        '${routerAddress}'
    ]

    def __init__(self, connector_type, config, i_am_request, reading_queue, rescan_queue, logger, converter_logger):
        self.__connector_type = connector_type
        self.__config = config
        DeviceObjectConfig.update_address_in_config_util(self.__config)
        self.alternative_responses_addresses = self.__config.get('altResponsesAddresses', [])

        self.__log = logger
        self.__converter_log = converter_logger

        self.__stopped = False
        self.active = True

        self.__request_process_queue = reading_queue
        self.__rescan_process_queue = rescan_queue

        if Device.need_to_retrieve_device_name(self.__config) and i_am_request.deviceName is None:
            self.__log.warning('Device name is not provided in IAmRequest. Device Id will be used as "objectName')
            i_am_request.deviceName = str(i_am_request.iAmDeviceIdentifier[1])

        self.details = BACnetDeviceDetails(i_am_request)
        self.device_info = DeviceInfo(self.__config.get('deviceInfo', {}), self.details)
        self.uplink_converter_config = UplinkConverterConfig(self.__config, self.device_info, self.details)

        self.name = self.device_info.device_name

        self.__objects_rescan_period = self.__config.get('devicesRescanObjectsPeriodSeconds', 60)
        self.rescan_objects_config = []

        self.__config_poll_period = self.__config.get('pollPeriod', 10000) / 1000
        self.__poll_period = self.__config_poll_period
        self.attributes_updates = self.__config.get('attributeUpdates', [])
        self.shared_attributes_keys = self.__get_shared_attributes_keys()
        self.server_side_rpc = self.__config.get('serverSideRpc', [])

        self.uplink_converter = self.__load_uplink_converter()

    def __str__(self):
        return f"Device(name={self.name}, address={self.details.address})"

    @property
    def config(self):
        return self.__config

    @property
    def stopped(self):
        return self.__stopped

    @property
    def original_poll_period(self):
        return self.__config_poll_period

    @property
    def poll_period(self):
        return self.__poll_period

    @poll_period.setter
    def poll_period(self, new_poll_period):
        self.__poll_period = new_poll_period

    @config.setter
    def config(self, new_config):
        self.__config = new_config
        self.uplink_converter_config = UplinkConverterConfig(new_config, self.device_info, self.details)
        self.uplink_converter = self.__load_uplink_converter()

    def __get_shared_attributes_keys(self):
        result = []

        for attr_config in self.attributes_updates:
            result.append(attr_config['key'])

        return result

    def __load_uplink_converter(self):
        try:
            if self.__config.get(UPLINK_PREFIX + CONVERTER_PARAMETER) is not None:
                converter_class = TBModuleLoader.import_module(self.__connector_type,
                                                               self.__config[UPLINK_PREFIX + CONVERTER_PARAMETER])
                converter = converter_class(self.uplink_converter_config, self.__converter_log)
            else:
                converter = AsyncBACnetUplinkConverter(self.uplink_converter_config, self.__converter_log)

            return converter
        except Exception as e:
            self.__log.exception('Failed to load uplink converter for % device: %s', self.name, e)

    def stop(self):
        self.active = False
        self.__stopped = True

    async def run(self):
        if len(self.uplink_converter_config.objects_to_read) > 0:
            self.__request_process_queue.put_nowait(self)

        next_poll_time = monotonic() + self.__poll_period

        while not self.__stopped:
            current_time = monotonic()
            if current_time >= next_poll_time:
                if len(self.uplink_converter_config.objects_to_read) > 0:
                    self.__request_process_queue.put_nowait(self)

                next_poll_time = current_time + self.__poll_period

            sleep_time = max(0.0, next_poll_time - current_time)

            await sleep(sleep_time)

    async def rescan(self):
        next_rescan_time = monotonic() + self.__objects_rescan_period

        while not self.__stopped:
            current_time = monotonic()
            if current_time >= next_rescan_time:
                if self.need_to_rescan():
                    self.__rescan_process_queue.put_nowait(self)

                next_rescan_time = current_time + self.__objects_rescan_period

            sleep_time = max(0.0, next_rescan_time - current_time)
            await sleep(sleep_time)

    @staticmethod
    def find_self_in_config(devices_config, apdu):
        apdu_address = apdu.pduSource.__str__()
        device_identifier = apdu.iAmDeviceIdentifier[-1]

        for device_config in devices_config:
            if device_config.get('deviceId') is not None:
                if Device.is_device_identifier_match(device_identifier, device_config.get('deviceId')):
                    return device_config
                else:
                    continue

            if Device.is_address_match(apdu_address, device_config.get('address')):
                return device_config
            elif apdu_address in device_config.get('altResponsesAddresses', []):
                return device_config

    @staticmethod
    def is_device_identifier_match(device_identifier, pattern):
        """
        Check if the device identifier matches the given pattern.
        :param device_identifier: The device identifier to check.
        :param pattern ("deviceId" in configuration file in device section):
        The pattern to match against. Can be a string, list, int or "*":
        - Asterisk (`*`) to match any device identifier.
        - A list of device identifiers to match against.
        - A string pattern that can include wildcards or regex patterns.
        - An integer to match a specific device identifier.
        """

        if pattern is not None:
            if pattern == '*':
                return True

            if isinstance(pattern, str):
                start, end = Device.__parse_range(pattern)
                return start <= device_identifier < end

            if isinstance(pattern, list):
                for rng in pattern:
                    start, end = Device.__parse_range(str(rng))
                    if start <= device_identifier < end:
                        return True

            return match(str(pattern), str(device_identifier)) is not None

        return False

    @staticmethod
    def is_address_match(address, pattern):
        regex = Device.get_address_regex(pattern)
        return match(regex, address) is not None

    @staticmethod
    def get_address_regex(initial_config_address):
        i = 0
        pattern = initial_config_address.split('@')[0]
        if pattern == '*:*:*':
            return r'.*'
        elif pattern == '*:*':
            return Device.ALLOWED_WITH_ROUTERS_ADDRESSES_PATTERN
        elif pattern == '*':
            return Device.ALLOWED_LOCAL_ADDRESSES_PATTERN_WITH_DEFAULT_PORT
        elif 'X' in pattern:
            port_present = False
            if pattern.count(':') == 1:
                port_present = True
                pattern_for_check = pattern.split(':')[0]
            else:
                pattern_for_check = pattern
            regex = ''
            while i < len(pattern_for_check):
                if pattern_for_check[i] == 'X':
                    while i < len(pattern_for_check) and pattern_for_check[i] == 'X':
                        i += 1
                    regex += r'\d+'
                else:
                    regex += escape(pattern_for_check[i])
                    i += 1
            if port_present:
                if pattern.split(':')[1] == '*':
                    regex += r'(:\d+)?'
                else:
                    if pattern.split(':')[1] == '47808':
                        regex += r'(?::47808)?'
                    else:
                        regex += r':'
                        regex += pattern.split(':')[1]
            return f"^{regex}$"
        elif ':' in pattern and '*' in pattern:
            if pattern.count(':') == 2:
                return Device.ALLOWED_WITH_ROUTERS_ADDRESSES_PATTERN
            elif pattern.count(':') == 1:
                if pattern.split(':')[1] == '47808':
                    return Device.ALLOWED_LOCAL_ADDRESSES_PATTERN_WITH_DEFAULT_PORT
                else:
                    expected_port = r'\d+' if pattern.split(':')[1] == '*' else pattern.split(':')[1]
                    return Device.ALLOWED_LOCAL_ADDRESSES_PATTERN_WITH_CUSTOM_PORT.replace('{expectedPort}', expected_port)  # noqa
        elif ':' in pattern:
            splitted = pattern.split(':')
            if len(splitted) > 1 and splitted[1] == "47808":
                return f"^{splitted[0]}(?::47808)?$"

        return f"^{pattern}$"

    @staticmethod
    def get_who_is_address(address):
        if 'X' in address or '*' in address:
            return Device.__get_broadcast_address(address)
        else:
            return address

    @staticmethod
    def __get_broadcast_address(address):
        parts = address.split(":")

        if len(parts) == 3:
            network = parts[0]
            return Device.__match_network(network)

        elif len(parts) == 2:
            if fullmatch(r"[0-9X]+", parts[0]):
                network = parts[0]
                return Device.__match_network(network)
            else:
                return "*:*"

        elif len(parts) == 1:
            return "*"

        return None

    @staticmethod
    def __match_network(network: str):
        if fullmatch(r"X+", network):
            return "*:*"
        else:
            return f"{network}:*"

    @staticmethod
    def is_pattern_address(address):
        return "X" in address

    @staticmethod
    def get_object_id(config):
        return ObjectIdentifier("%s:%s" % (config['objectType'], config['objectId']))

    @staticmethod
    def is_global_discovery_config(config):
        fill_for = []

        if config.get('attributes', '') == '*':
            fill_for.append('attributes')

        if config.get('timeseries', '') == '*':
            fill_for.append('timeseries')

        return fill_for

    @staticmethod
    def is_local_discovery_config(config):
        fill_for = []

        for section in ('attributes', 'timeseries'):
            for item_config in config.get(section, []):
                if isinstance(item_config, dict) and item_config['objectId'] == '*':
                    item_config['type'] = section
                    fill_for.append(item_config)

        return fill_for

    @staticmethod
    def need_to_retrieve_device_name(config):
        if '${objectName}' in config.get('deviceInfo', {}).get('deviceNameExpression', ''):
            return True
        if '${objectName}' in config.get('deviceInfo', {}).get('deviceProfileExpression', ''):
            return True

        return False

    @staticmethod
    def need_to_retrieve_router_info(config):
        for param in Device.ROUTER_INFO_PARAMS:
            if param in config.get('deviceInfo', {}).get('deviceNameExpression', ''):
                return True
            if param in config.get('deviceInfo', {}).get('deviceProfileExpression', ''):
                return True

        return False

    @staticmethod
    def parse_config_key(config):
        if '${' in config['key']:
            result_tags = set()
            result_tags.update(TBUtility.get_values(config['key'], {}, get_tag=True))
            if len(result_tags) > 0:
                if config['propertyId'] != '*':
                    if isinstance(config['propertyId'], list):
                        result_tags.update(config['propertyId'])
                        config['propertyId'] = result_tags
                    elif isinstance(config['propertyId'], str):
                        result_tags.add(config['propertyId'])
                        config['propertyId'] = result_tags

    @staticmethod
    def parse_ranges(object_id_config):
        ranges = []

        if isinstance(object_id_config, list):
            for rng in object_id_config:
                start, end = Device.__parse_range(rng)
                ranges.append((start, end))
        elif isinstance(object_id_config, str):
            if object_id_config != '*':
                start, end = Device.__parse_range(object_id_config)
                ranges.append((start, end))

        return ranges

    @staticmethod
    def __parse_range(rng):
        rng = rng.strip()
        if '-' in rng:
            start, end = map(str.strip, rng.split('-'))
            return int(start), int(end) + 1
        else:
            return int(rng), int(rng) + 1

    def need_to_rescan(self):
        return len(self.details.failed_to_read_indexes) > 0 and not self.details.is_segmentation_supported()


class Devices:
    def __init__(self):
        self.__devices = {}
        self.__devices_by_name = {}
        self.__lock = Lock()

    async def add(self, device):
        if not isinstance(device, Device):
            raise TypeError("Expected a Device instance")

        await self.__lock.acquire()
        try:
            self.__devices[device.details.object_id] = device
            self.__devices_by_name[device.name] = device
        finally:
            self.__lock.release()

    async def remove(self, device):
        await self.__lock.acquire()
        try:
            self.__devices.pop(device.details.object_id, None)
            self.__devices.pop(device.name, None)
        finally:
            self.__lock.release()

    async def get_device_by_id(self, _id):
        await self.__lock.acquire()
        try:
            item = self.__devices.get(_id)
        finally:
            self.__lock.release()

        return item

    async def get_device_by_name(self, name):
        await self.__lock.acquire()
        try:
            item = self.__devices_by_name.get(name)
        finally:
            self.__lock.release()

        return item

    def stop_all(self):
        for device in self.__devices.values():
            device.stop()



from pymodbus.constants import Endian


class BytesDownlinkConverterConfig:
    def __init__(self, device_name, byte_order, word_order, repack, objects_count, function_code, lower_type, address):
        self.device_name = device_name
        self.byte_order = Endian.BIG if byte_order.upper() == "BIG" else Endian.LITTLE
        self.word_order = Endian.BIG if word_order.upper() == "BIG" else Endian.LITTLE
        self.repack = repack
        self.objects_count = objects_count
        self.function_code = function_code
        self.lower_type = lower_type.lower()
        self.address = address



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import time
from os import path, listdir


class Pointer:
    def __init__(self, path_to_folder, log):
        self.__log = log
        self.__directory = path.dirname(path_to_folder)

    def sort_db_files(self):
        all_files = listdir(self.__directory)
        all_db_files = sorted(filter(lambda file: file.endswith(".db"), all_files))
        return all_db_files

    @staticmethod
    def generate_new_file_name():
        prefix = "data_"
        ts = str(int(time.time()) * 1000)
        db_file_name = f"{prefix}{ts}.db"
        return db_file_name
